"""A module containing all credential types"""
from azol.credentials.credential import Credential
from azol.credentials.access_token import AccessToken
from azol.credentials.service_principal import ServicePrincipal
from azol.credentials.application_object import ApplicationObject
from azol.credentials.user import User

__all__ = [
    "AccessToken",
    "ServicePrincipal",
    "ApplicationObject",
    "User"
]



"""Module containing all token caches for azol"""
from azol.caches.in_memory_token_cache import InMemoryTokenCache
from azol.caches.local_token_cache import LocalTokenCache

__all__ = [
    "InMemoryTokenCache",
    "LocalTokenCache"
]



"""Module containing the DataFactoryClient class"""
import uuid
import logging
import requests
import base64
import json
from time import sleep

class TooManyIRNodesException(Exception):
    pass

class DataFactoryClient:
    """
        An HTTP Client for the Internal DataFactory API
    """

    def __init__( self, authentication_key, ir_node_id=None, spoofed_ir_name="Default" ):
        self.key=authentication_key
        self.dataFactory_hostname=None
        self.dataFactory_url=None
        self.user_key_encoded=None
        self.ir_instance_id=None
        self.dataFactory_name=None
        self.datafactory_token=None
        self.subscription_id=None
        self.data_factory_id=None
        self.current_bearer_token=None
        self.ir_node_id=ir_node_id
        self.spoofed_ir_name=spoofed_ir_name

        # Parse the key and construct the Base64 encoded token that will be used later
        (_, self.ir_instance_id, self.dataFactory_name, service_endpoint, ir_key)=self.key.split("@")
        self.dataFactory_hostname=service_endpoint.split("=")[1]
        self.dataFactory_url= f"https://{self.dataFactory_hostname}"
        user_key_unencoded=f"ams#{self.dataFactory_name}#{self.ir_instance_id}#{ir_key}".encode()
        self.user_key_encoded=base64.urlsafe_b64encode(user_key_unencoded).decode()

        self._get_and_parse_df_bearer()
    
    def _get_and_parse_df_bearer(self):
        bearer_token=self.get_dataFactory_token()
        self.current_bearer_token=bearer_token

        # Decode the token and extract the Azure subscription ID of the Data Factory
        decoded_df_token = base64.b64decode(bearer_token)
        df_token_dict = json.loads(decoded_df_token)
        df_token_content=json.loads(df_token_dict["ContentString"])
        declarations_dict = json.loads(df_token_content["Policy"]["Declarations"][0]["Value"])
        self.subscription_id=declarations_dict["PartitionId"]
        self.data_factory_id=declarations_dict["ExtendedProperties"]["DataFactoryId"]
        
    def get_dataFactory_token(self):
        headers = {
            "Authorization": f"UserKey {self.user_key_encoded}",
            "Content-Type": "application/json"
        }

        response = requests.post(self.dataFactory_url + "/subscriptions/00000000-0000-0000-0000-000000000000/authservice/ams/acquiretoken",
                         headers=headers, json=self.user_key_encoded)
        
        data_factory_token = response.json()
        return data_factory_token

    def get_managed_identity_token(self, oauth_resource):
        """
            Get a managed identity token using the DataFactory key
            
            Args:
            - oauth_resource - (string) The oauth resource for the managed identity request.
                                           f.eks https://management.azure.com/, or the client_id
                                           of a service principal

            Returns:
                (string) The access token received
        """
        token=self.get_dataFactory_token()
        headers = {
            "Authorization": f"Bearer {token}"
        }

        body = {
            "resource": oauth_resource,
            "entityName": self.dataFactory_name,
            "executionContextRestricted": True,
            "executionContextVersion": "2024-01-01"
        }

        query_params = {
            "api-version": "2019-09-01"
        }

        request_url = f"https://{self.dataFactory_hostname}/subscriptions/00000000-0000-0000-0000-000000000000/entities/{self.data_factory_id}/identities/00000000-0000-0000-0000-000000000000/token"
        response=requests.post(request_url, headers=headers, params=query_params, json=body)
        token_response=response.json()
        return token_response["accessToken"]

    def spoof_shir(self, poll_interval=3, job_callback=None):
        """
            Register a new self-hosted integration runtime(shir) with the data factory using the
            authentication key. Poll every 3 seconds for jobs. If a job comes in, either print it
            to screen, or process it with the job_callback.

            Args:
            - poll_interval: (int) The frequency to poll for new jobs
            - job_callback: (func) A callback function to execute whenever a new job is observed.
                                   This is useful for custom processing of job responses.

                                   callback arguments: function(<dict>job)
        """
        token=self.get_dataFactory_token()
        auth_headers = {
            "Authorization": f"Bearer {token}",
            "X-Ms-Gateway-Machine-Id": str(uuid.uuid4())
        }
        if self.ir_node_id is None:
            # Register a new Node and get the ID

            body = {
                "gatewayVersion": {
                    "major":5,
                    "minor":48,
                    "build":9076,
                    "revision":1,
                    "majorRevision":0,
                    "minorRevision":1
                },
                "machineName": self.spoofed_ir_name,
                "nodeName":self.spoofed_ir_name
            }

            response = requests.post(f"{self.dataFactory_url}/tenants/{self.subscription_id}/gateways/{self.ir_instance_id}/nodes?api-version=1.0",
                            json=body, headers=auth_headers)
            
            node_response=response.json()
            if "errorCode" in node_response:
                logging.error(f"Error while creating node: {node_response['errorCodeString']}")
                raise TooManyIRNodesException
            self.ir_node_id=node_response["instanceId"]
            logging.info(f"New SHIR Node Registered Data Factory {self.dataFactory_name}: {self.ir_node_id}")
        
        # Establish a new connection
        body = {
            "gatewayVersion": {
                "major":5,
                "minor":48,
                "build":9076,
                "revision":1,
                "majorRevision":0,
                "minorRevision":1
            },
            "machineName": self.spoofed_ir_name
        }

        response = requests.post(f"{self.dataFactory_url}/tenants/{self.subscription_id}/gateways/{self.ir_instance_id}/nodes/{self.ir_node_id}/establishconnection?api-version=1.0",
                                json=body, headers=auth_headers)
        connection_info = response.json()
        public_key=connection_info["publicKey"]

        # Report that we are connected
        body = {
            "serviceBusConnected": None,
            "httpsPortEnabled": True,
            "workerActivated": True
        }

        response = requests.post(f"{self.dataFactory_url}/tenants/{self.subscription_id}/gateways/{self.ir_instance_id}/nodes/{self.ir_node_id}/reportstatus?api-version=1.0",
                         json=body, headers=auth_headers)

        # send initialize request
        body = {
                "gatewayVersion": {
                "major":5,
                "minor":48,
                "build":9076,
                "revision":1,
                "majorRevision":0,
                "minorRevision":1
            },
            "nodeName": self.spoofed_ir_name,
            "publicKey": public_key,
            "coreInitialized":True,

        }

        response = requests.post(f"{self.dataFactory_url}/tenants/{self.subscription_id}/gateways/{self.ir_instance_id}/nodes/{self.ir_node_id}/initialize?api-version=1.0",
                         json=body, headers=auth_headers)
        
        # poll for jobs and print out the contents of the request if it is found, or process it with the callback.
        while True:
            logging.info("Polling for jobs...")
            token = self.get_dataFactory_token()
            auth_headers = {
                "Authorization": f"Bearer {token}"
            }

            query_params = {
                "issecondary": False,
                "queueid": self.ir_instance_id,
                "count": 1,
                "workerid": uuid.uuid4(),
                "api-version": "2014-06-26"
            }

            response = requests.get(f"{self.dataFactory_url}/subscriptions/{self.subscription_id}/we/tasks/next",
                                    headers=auth_headers, params=query_params)
            if response.status_code == 204:
                pass
            else:
                logging.info("Got Job.")
                response_content = response.json()
                for job in response_content:
                    input_data = json.loads(job["InputData"])

                    properties_dict=input_data["Properties"]
                    logging.debug("Job Properties:")
                    logging.debug(properties_dict)
                    if job_callback is None:
                        logging.info("Printing Job Output to stdout...")
                        print(properties_dict)
                    else:
                        logging.info("Using Callback...")
                        job_callback(properties_dict)
            sleep(poll_interval)


"""Module containing a token cache that persists on the local filesystem"""
import json
import sys
from pathlib import Path
import logging

from azol.constants import TOKENCACHEDIR
from azol.caches.azol_cache import AzolCache

class LocalTokenCache( AzolCache ):
    """class for an azol token cache that persists on the local filesystem"""

    def __init__( self, *args, **kwargs ):
        super().__init__( *args, **kwargs)
        Path(TOKENCACHEDIR).mkdir(parents=True, exist_ok=True)
        self.file_name=TOKENCACHEDIR+"/default"
        Path(self.file_name).touch(exist_ok=True)

    def try_get_token(self, *args, **kwargs):
        """
            Attempt to get a token from the cache that matches the requested tenantId,
            clientId, scope, and user.
        """

        with open(self.file_name, "r", encoding="utf-8") as f:
            try:
                self.tokens=json.load(fp=f)
            except json.decoder.JSONDecodeError:
                logging.info("cannot load token cache as %s because it is an empty file. "
                             "SNo tokens saved", self.file_name)
                return None

        return super().try_get_token(*args, **kwargs)

    def cache_or_update(self, *args, **kwargs):
        """
            Saves an access and refresh token to the cache, or updates an existing entry.
            Overwrites the existing local cache file
        """

        #read configuration from file
        try:
            with open(self.file_name, "r", encoding="utf-8") as f:
                self.tokens=json.load(fp=f)
        except json.decoder.JSONDecodeError as msg:
            data=""
            with open(self.file_name, "r", encoding="utf-8") as f:
                data = f.read()
            if data == "":
                logging.warning("Token cache was empty. initializing")
                with open(self.file_name, "w", encoding="utf-8") as f:
                    json.dump({}, fp=f)
                self.tokens={}
            else:
                logging.error( "fatal error: %s", msg)
                sys.exit()

        super().cache_or_update(*args, **kwargs)

        #re-write the new tokenCache to the file
        with open(self.file_name, "w", encoding="utf-8") as f:
            json.dump(self.tokens, indent=4, sort_keys=True, fp=f)



"""Module that contains functions that are useful when working with Entra ID and Azure"""
import logging
import json
import base64
import time
import requests

def string_between(whole_string, start_string, end_string, include_start=False):
    i=whole_string.index(start_string)
    z=whole_string[i:].index(end_string)+i
    if not include_start:
        i=i+len(start_string)
    sub_string=whole_string[i:z]
    return sub_string

def get_tenant_id( tenant_name ):
    """
        Call the openid-configuration of a domain name to check if it exists in Entra ID as a tenant

        Returns: (string) tenant ID if the domain name is registered, or None
    """
    response = requests.get(f"https://login.microsoftonline.com/{tenant_name}/"
                            "v2.0/.well-known/openid-configuration", timeout=10)

    if response.status_code == 400:
        logging.info("domain name not registered as a tenant")
        return None
    well_known=response.json()
    issuer=well_known['token_endpoint']
    tenant_id = issuer.split('/')[3]
    return tenant_id

def parse_jwt( token ):
    """
        parse the jwt and return an object for each piece

        Args:
            - token - (string) The raw ascii token to parse

        Returns: a tuple ( <dict>header, <dict>data, <string>signature )
    """
    def fix_padding( raw_string ):
        while True:
            if len(raw_string) % 4 == 0:
                break
            raw_string += "="
        return raw_string
    if token is None:
        return True
    header, data, signature = token.split(".")
    headerb64 = fix_padding( header )
    datab64 = fix_padding( data )
    signatureb64 = fix_padding( signature )
    header = json.loads(base64.b64decode(headerb64).decode())
    data = json.loads(base64.b64decode(datab64).decode())
    return header, data, signatureb64

def is_token_expired( token, padding_time=0 ):
    """
        Check if the current access token is expired. The credential
        object caches access tokens in memory. This method will check
        if the current cached token is expired.

        Args:
            - paddingTime - (int) Check if token expires, [paddingTime]
                                  seconds before it actually expires.
                                  Useful if you want to get a new token
                                  prior to expiry by some margin.
        
        Returns: True if expired, False if not expired
        
    """
    _, data, _ = parse_jwt(token)
    expiry = data[ "exp" ]
    current_time = int( time.time() )
    if current_time >= expiry-padding_time:
        return True
    return False



"""
    Module containing the token service, which is used for fetching tokens
    for the azol entra ID OAuth clients
"""
import logging
import uuid
import json
from time import sleep
import requests
import getpass
import urllib.parse
from html.parser import HTMLParser
from azol.constants import IDENTITYPLATFORMURL, DEFAULTSCOPE
from azol.utils import is_token_expired, parse_jwt, string_between
from azol.caches import LocalTokenCache, InMemoryTokenCache
from azol.constants import OAUTHFLOWS, FOCIClients, known_client_redirect_uris
from copy import deepcopy

from .token_service_helpers import build_scope_string, auth_code_flow, ests_login_flow

class IdentityPlatformRequestFailedException(Exception):
    """
        Exception that is raised when requests to the identity platform
        unexpectedly fail
    """

class TokenService( object ):
    """
        Generic OAuth2 token service for Entra ID
    """

    def __init__( self, cred, tenant_id, oauth_flow,
                  secrets_provider, use_persistent_cache, oauth_resource,
                  scopes, default_scope, profile_scope, openid_scope,
                  offline_access_scope, azol_id=None):

        self.credential_object=deepcopy(cred)
        self._tenant=tenant_id
        self.oauth_flow=oauth_flow
        self.secrets_provider=secrets_provider
        self.oauth_resource=oauth_resource
        self.scopes=scopes
        self.default_scope=default_scope
        self.profile_scope=profile_scope
        self.openid_scope=openid_scope
        self.offline_access_scope=offline_access_scope
        self.ests_cookie=None
        self.ests_persistent_cookie=None
        if self.credential_object.has_ests():
            cookies=self.credential_object.get_ests()
            self.ests_cookie=cookies.ests
            self.ests_persistent_cookie=cookies.ests_persistent
        else:
            self.ests_cookie=None
            self.ests_persistent_cookie=None
        if self.credential_object.has_refresh_token():
            self._refresh_token=self.credential_object.get_refresh_token()
        else:
            self._refresh_token=None

        if self.credential_object.credentialType == "user":
            self.flow_type="delegated"
        else:
            self.flow_type="application"

        self._client_id=self.credential_object.get_client_id()

        self.registered_token_flows={}
        self._register_flow( OAUTHFLOWS.CLIENT_CREDENTIALS, self._get_client_credential_token )
        self._register_flow( OAUTHFLOWS.REFRESH_TOKEN , self._refresh_token_flow )
        self._register_flow( OAUTHFLOWS.DEVICE_CODE, self._device_code )
        self._register_flow( OAUTHFLOWS.RAW_TOKEN, self._raw_token )
        self._register_flow( OAUTHFLOWS.AUTHORIZATION_CODE, self._authorization_code)
        if azol_id is None:
            azol_id = uuid.uuid4()
        self._id = azol_id

        if use_persistent_cache:
            self.token_cache=LocalTokenCache()
        else:
            self.token_cache=InMemoryTokenCache()

        if ( OAUTHFLOWS.DEVICE_CODE in self.credential_object.supportedOAuthFlows 
             or OAUTHFLOWS.REFRESH_TOKEN in self.credential_object.supportedOAuthFlows
              or OAUTHFLOWS.AUTHORIZATION_CODE in self.credential_object.supportedOAuthFlows ):
            get_cached_token_result = self.token_cache.try_get_token(self._tenant,
                                                            self._client_id,
                                                            self.default_scope, self.scopes,
                                                            self.oauth_resource,
                                                            self.credential_object._username )
            if get_cached_token_result is not None:
                refresh_token=get_cached_token_result["refresh_token"]
                ests=get_cached_token_result["ests_cookie"]
                ests_persistent_cookie=get_cached_token_result["ests_persistent_cookie"]
                self._refresh_token=refresh_token
                self.ests_cookie=ests
                self.ests_persistent_cookie=ests_persistent_cookie
                self.oauth_flow=OAUTHFLOWS.REFRESH_TOKEN
            else:
                logging.info( "Tried to use a token from the cache but none exists" )

    def set_tenant( self, tenant ):
        self._tenant=tenant

    def set_scope(self, scopes, default, profile, offline_access, openid):
        self.scopes=scopes
        self.default_scope=default
        self.profile_scope=profile
        self.offline_access_scope=offline_access
        self.openid_scope=openid

    def switch_client( self, client_id ):
        self._client_id=client_id

    def refresh_token( self ):
        if self.credential_object.credentialType != "user":
            raise Exception("Cannot refresh a token for a non user credential type")
        if self._refresh_token == None:
            raise Exception('Cannot refresh token - no refresh token present to refresh')

        token_data_object=self._refresh_token_flow()
        refresh_token = token_data_object["refresh_token"]
        access_token=token_data_object["access_token"]
        self._refresh_token=refresh_token

        self.token_cache.cache_or_update( access_token, self._tenant,
                                         self._client_id,
                                         self.default_scope, self.scopes, self.oauth_resource,
                                         refresh_token=refresh_token,
                                         username=self.credential_object.get_username() )
        return access_token

    def get_refresh_token_if_exists( self ):
        """
            Returns the current refresh token saved by the token service

            Returns: string - raw refresh token if it is saved, else None
        """
        if self.credential_object.credentialType != "user":
            raise Exception("Cannot use a refresh token for a non user credential type")

        if self._refresh_token is not None:
            return self._refresh_token
        return None

    def get_id( self ):
        """
            Get the azol Id of the tokenService object

            Returns string - azol id
        """
        return self._id

    def _register_flow( self, flow_name, flow_function ):
        """
            Internal method. register an oauth flow function with the tokenService tokenFlows map

            Args:
                - flow_name - string - the azol name of the oauth flow to call
                - flow_function -function - the function that
                                  should be added to the registered flows map

            Returns None
        """
        self.registered_token_flows[ flow_name ] = flow_function

    def get_cached_token_if_not_expired( self ):
        """
            Get a cached token from the token Cache if it exists

            Returns: string of ascii encoded token, or None if no token found in cache
        """
        token_data=self.token_cache.try_get_token( self._tenant,
                                                  self._client_id,
                                                  self.default_scope, self.scopes,
                                                  self.oauth_resource,
                                                  self.credential_object.get_username() )

        if token_data is None:
            logging.info("get_cached_token_if_not_expired: Token does not exist in cache")
            return None
        token = token_data["access_token"]
        if is_token_expired( token ):
            return None
        return token

    def fetch_token( self ):
        """
            Initiates an authorization flow and caches a new token.

            Returns: None if a token couldnt be fetched, or the access token
        """
        get_token_function = self.registered_token_flows[self.oauth_flow]
        token_data_object = get_token_function()
        if token_data_object is None:
            logging.error("Could not get a new token")
            raise IdentityPlatformRequestFailedException()
        access_token = token_data_object["access_token"]
        ests_cookie=None
        ests_persistent_cookie=None
        if "ests" in token_data_object.keys():
            ests_cookie=token_data_object["ests"]
        if "estspersistent" in token_data_object.keys():
            ests_persistent_cookie=token_data_object["estspersistent"]

        if self._tenant=="common":
            _, body, _ = parse_jwt(access_token)
            self.set_tenant(body["tid"])
            
        if self.offline_access_scope and self.credential_object.credentialType=="user":
            refresh_token = token_data_object["refresh_token"]
            self.token_cache.cache_or_update( access_token, self._tenant,
                                            self._client_id,
                                            self.default_scope, self.scopes, self.oauth_resource,
                                            refresh_token=refresh_token,
                                            username=self.credential_object.get_username(),
                                            ests_cookie=ests_cookie,
                                            ests_persistent_cookie=ests_persistent_cookie )
        else:
            self.token_cache.cache_or_update( access_token, self._tenant,
                                            self._client_id,
                                            self.default_scope, self.scopes, self.oauth_resource,
                                            ests_cookie=ests_cookie,
                                            ests_persistent_cookie=ests_persistent_cookie )
        return access_token

    def get_cached_token( self ):
        """
            Get a cached token from the token Cache if it exists.
            Will also return expired tokens.

            Returns: string of ascii encoded token, or None if no token found in cache
        """
        token_data=self.token_cache.try_get_token( self._tenant,
                                                self._client_id,
                                                self.default_scope, self.scopes,
                                                self.oauth_resource,
                                                self.credential_object.get_username() )

        if token_data is None:
            return None

        token = token_data["access_token"]
        return token

    def get_client_id(self):
        return self._client_id

    def _raw_token( self ):
        """
            registered when a raw token is used as a credential -
                dont perform any oauth flow, just return the raw token

            :returns dictionary containing token data
        """
        if OAUTHFLOWS.RAW_TOKEN not in self.credential_object.supportedOAuthFlows:
            logging.error( "credential type of %s does not support "
                          "login flow of type raw_token.", 
                          self.credential_object )
            raise Exception( "Unsupported Flow" )
        credential_value = self.credential_object.token
        if credential_value.startswith("secret:"):
            if self.secrets_provider:
                return self.secrets_provider.getSecret(credential_value.split("secret:")[1])
            else:
                logging.error( "critical error: secret referenced used "
                              "for raw_token credential, but no secret provider configured. "
                              "exiting..." )
                exit()
        new_token_data = {
            "access_token": self.credential_object.token,
            "refresh_token": None,
            "flow_type": "unknown"
        }
        return new_token_data

    def _get_client_credential_token( self ):
        """
            client credential grant

            Returns: dictionary containing token data
        """
        if OAUTHFLOWS.CLIENT_CREDENTIALS not in self.credential_object.supportedOAuthFlows:
            logging.error( "credential type of %s does not"
                          " support OAuth flow of type client_credential.",
                          self.credential_object )
            raise Exception( "Unsupported OAuth Flow" )

        url = f"{IDENTITYPLATFORMURL}/{self._tenant}/oauth2/v2.0/token"

        secret_from_credential = self.credential_object.get_client_secret()
        if secret_from_credential.startswith("secret:"):
            if self.secrets_provider:
                client_secret = self.secrets_provider.get_secret(
                       secret_from_credential.split("secret:")[1] )
            else:
                logging.error( "critical error: secret referenced used for servicePrincipal "
                              "credential, but no secret provider configured. exiting..." )
                raise Exception("secret refrerence used for sp credential, "
                                "but there no secret provider")
        else:
            client_secret = secret_from_credential

        openid_scope = "openid"
        profile_scope = "profile"
        offline_access_scope = "offline_access"
        extension = ""
        if self.openid_scope:
            extension+=f" {openid_scope}"
        if self.profile_scope:
            extension+=f" {profile_scope}"
        if self.offline_access_scope:
            extension+=f" {offline_access_scope}"

        if self.default_scope:
            scope_string = f"{self.oauth_resource}{DEFAULTSCOPE}"
            extended_scope_string=scope_string+extension
        else:
            expanded_scopes = [ f"{self.oauth_resource}{s}" for s in self.scopes ]
            scope_string = " ".join(expanded_scopes)
            extended_scope_string=scope_string+extension
        body = {
            "client_id": self._client_id,
            "client_secret": client_secret,
            "scope": extended_scope_string,
            "grant_type": OAUTHFLOWS.CLIENT_CREDENTIALS
        }

        response = requests.post( url, data=body, timeout=10 )
        response_status = response.status_code
        logging.info( "response status for client credential grant: %s", str(response_status) )
        if response_status != 200:
            error_msg = response.content
            logging.error( "Client credential grant failed for client id "
                          "%s at scope %s."
                          " Raw Identity Platform error message:\r\n%s", 
                          self._client_id, extended_scope_string,
                          error_msg )
            return
        response_body = response.json()
        token = response_body[ "access_token" ]
        new_token_data = {
            "access_token": token,
            "flow_type": "application"
        }
        return new_token_data

    def _refresh_token_flow( self ):
        """
            Refresh token flow

            Returns: dictionary containing token data
        """
        if OAUTHFLOWS.REFRESH_TOKEN not in self.credential_object.supportedOAuthFlows:
            logging.error( "credential type of %s does not support OAuth flow of "
                          "type refresh_token.", self.credential_object )
            raise Exception( "Unsupported OAuth Flow" )

        client_id=self._client_id

        openid_scope = "openid"
        profile_scope = "profile"
        offline_access_scope = "offline_access"
        extension = ""
        if self.openid_scope:
            extension+=f" {openid_scope}"
        if self.profile_scope:
            extension+=f" {profile_scope}"
        if self.offline_access_scope:
            extension+=f" {offline_access_scope}"

        if self.default_scope:
            scope_string = f"{self.oauth_resource}{DEFAULTSCOPE}"
            extended_scope_string=scope_string+extension
        else:
            expanded_scopes = [ f"{self.oauth_resource}{s}" for s in self.scopes ]
            scope_string = " ".join(expanded_scopes)
            extended_scope_string=scope_string+extension

        body = {
            "client_id" : client_id,
            "refresh_token": self._refresh_token,
            "scope": extended_scope_string,
            "grant_type": "refresh_token"
        }

        response = requests.post( "https://login.microsoftonline.com/"
                                 f"{self._tenant}/oauth2/v2.0/token", data=body,  timeout=10 )
        if "error" in response.json().keys():
            error_msg = response.content
            msg_dict = json.loads(error_msg)
            eid_error_code=msg_dict["error_description"].split(":")[0]
            if eid_error_code=="AADSTS50076":
                # We need to use MFA to switch tenants.
                # This requires an ESTSAUTH or ESTSPERSISTENTAUTH cookie
                if self.ests_cookie!=None:
                    # Use the ests cookie to start MFA again
                    new_token_data=ests_login_flow(self._tenant, self._client_id,
                                                   self.ests_cookie, self.ests_persistent_cookie, 
                                                   self.credential_object.get_username(),
                                                   extended_scope_string, self.credential_object._redirect_uri)
                    self._refresh_token=new_token_data["refresh_token"]
                    return new_token_data
            else:
                logging.error( "Error: error while refreshing token. "
                            "Raw Error Message from Identity Platform: %s", error_msg )
                
                raise IdentityPlatformRequestFailedException()
            
        if "refresh_token" in response.json().keys():
            token = response.json()[ "access_token" ]
            refresh_token = response.json()[ "refresh_token" ]

            # It is possible that we dont know whos refresh token this is.
            # Extract the username and set it in the credential object
            if not self.credential_object.username_is_known():
                _, body, _ = parse_jwt(token)
                self.credential_object.set_username(body["unique_name"])
            new_token_data = {
                "access_token": token,
                "refresh_token": refresh_token,
                "flow_type": "delegated"
            }
            return new_token_data

    def _authorization_code( self ):
        """
            authorization code flow

            This flow defaults to an idp-initiated login flow.
            
            If a client is used that does not
            support idp-initiated flows, it will fail by default. However, some clients such as 
            Azure Portal are supported by Azol, and Azol will interrupt the login to force an
            SP-initiated flow. The supported clients that only support SP intiated flows include:

              - Azure Portal

            Returns: dictionary containing token data
        """
        if OAUTHFLOWS.AUTHORIZATION_CODE not in self.credential_object.supportedOAuthFlows:
            logging.error( "credential type of %s does not support OAuth flow of "
                          "type authorization_code.", self.credential_object )
            raise Exception( "Credential does not support OAuth Flow" )
        
        client_id=self._client_id

        redirect_url=self.credential_object._redirect_uri

        scope_string=build_scope_string(self.oauth_resource, self.scopes, self.default_scope,
                                        self.openid_scope, self.profile_scope, self.offline_access_scope )
        
        new_token_data=auth_code_flow(self._tenant, self.credential_object.get_username(), client_id, redirect_url, scope_string)
        
        self.ests_cookie=new_token_data["ests"]
        self.ests_persistent_cookie=new_token_data["estspersistent"]

        self._refresh_token=new_token_data["refresh_token"]
        # start refreshing tokens in the http client instead of re-authenticating
        self.oauth_flow=OAUTHFLOWS.REFRESH_TOKEN

        return new_token_data

    def _device_code( self ):
        """
            device code flow

            Returns: dictionary containing token data
        """

        client_id = self._client_id
        if OAUTHFLOWS.DEVICE_CODE not in self.credential_object.supportedOAuthFlows:
            logging.error( "credential type of %s does not support "
                          "OAuth flow of type device_code.", self.credential_object.__name__ )
            raise Exception( "Unsupported OAuth Flow" )
        openid_scope = "openid"
        profile_scope = "profile"
        offline_access_scope = "offline_access"
        extension = ""
        if self.openid_scope:
            extension+=f" {openid_scope}"
        if self.profile_scope:
            extension+=f" {profile_scope}"
        if self.offline_access_scope:
            extension+=f" {offline_access_scope}"

        if self.default_scope:
            scope_string = f"{self.oauth_resource}{DEFAULTSCOPE}"
            extended_scope_string=scope_string+extension
        else:
            expanded_scopes = [ f"{self.oauth_resource}{s}" for s in self.scopes ]
            scope_string = " ".join(expanded_scopes)
            extended_scope_string=scope_string+extension
        body = {
            "client_id": client_id,
            "scope": extended_scope_string, 
        }
        response = requests.post( "https://login.microsoftonline.com/"
                                 f"{self._tenant}/oauth2/v2.0/devicecode",
                                 data=body, timeout=10 )
        if response.status_code != 200:
            logging.error("Error while getting a device code: %s", response.content)
            raise IdentityPlatformRequestFailedException()
        auth_url = response.json()[ "verification_uri" ]
        user_code = response.json()[ "user_code" ]
        device_code = response.json()[ "device_code" ]
        print( f"Visit site {auth_url} from a browser, and enter code {user_code}" )
        timeout = 24
        retry_number = 0
        while True:
            sleep( 5 )
            logging.info( "Polling device code endpoint for auth token..." )
            body = {
                "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                "client_id": client_id,
                "device_code": device_code
            }
            response = requests.post( "https://login.microsoftonline.com/"
                                     f"{self._tenant}/oauth2/v2.0/token",
                                     data=body, timeout=10 )
            if response.status_code != 200:

                if response.json()[ "error" ] != "authorization_pending":
                    logging.error( "error while logging in via device code flow. description: %s",
                                  response.json()[ 'error_description' ] )
                    logging.error( response.json()[ "error_description" ] )
                    raise Exception("Error getting token during device_code flow.")
                logging.info("Authorization pending, continuing to poll...")
                retry_number = retry_number + 1
                if retry_number == timeout:
                    logging.error( "timeout on device code flow. token not retrieved" )
                    raise Exception("Error getting token during device_code flow.")
                continue

            token = response.json()[ "access_token" ]
            refresh_token = response.json()[ "refresh_token" ]

            # check if the user logged in with the user that they specified
            _, body, _ = parse_jwt(token)
            username=body["unique_name"].lower()

            # We need to know the username at startup so that we can check for cached tokens
            if username != self.credential_object.get_username():
                logging.error("Please log in with the same user specified in the User() object." )
                raise Exception("username of credential object does not "
                                "match username of logged-in user")

            self._refresh_token=refresh_token

            # start refreshing tokens from now on
            self.oauth_flow=OAUTHFLOWS.REFRESH_TOKEN

            # We dont know what user the user logged in with at this point.
            # Extract the username from the token
            _, body, _ = parse_jwt(token)
            logging.info( "Retrieved refresh token for the user" )
            break

        new_token_data = {
            "access_token": token,
            "refresh_token": refresh_token,
            "flow_type": "delegated"
        }
        return new_token_data



"""A module containing an azure generic resource model"""
from dataclasses import dataclass
from typing import Optional

@dataclass
class GenericResource( object ):
    """
        A generic resource model that represents azure objects
    """
    id: str
    name: str
    properties: object
    type: str
    changedTime: Optional[str] = None
    createdTime: Optional[str] = None
    extendedLocation: Optional[object] = None
    identity: Optional[object] = None
    kind: Optional[str] = None
    location: Optional[str] = None
    managedBy: Optional[str] = None
    plan: Optional[object] = None
    provisioningState: Optional[str] = None
    sku: Optional[object] = None
    tags: Optional[object] = None
    azolAnnotations : Optional[object] = None

    


import logging
import uuid
import json
from time import sleep
import requests
import getpass
import urllib.parse
import string
import random
from html.parser import HTMLParser
from azol.constants import IDENTITYPLATFORMURL, DEFAULTSCOPE
from azol.utils import is_token_expired, parse_jwt, string_between
from azol.caches import LocalTokenCache, InMemoryTokenCache
from azol.constants import OAUTHFLOWS, FOCIClients, known_client_redirect_uris
from copy import deepcopy
from azol.utils.auth_utils import start_azure_portal_login, end_azure_portal_login

def build_scope_string(oauth_resource, scopes=[], default_scope=None, openid_scope=None,
                       profile_scope=None, offline_access_scope=None ):
    extension = ""
    if openid_scope:
        extension+=f" openid"
    if profile_scope:
        extension+=f" profile"
    if offline_access_scope:
        extension+=f" offline_access"

    if default_scope:
        scope_string = f"{oauth_resource}{DEFAULTSCOPE}"
        extended_scope_string=scope_string+extension
    else:
        expanded_scopes = [ f"{oauth_resource}{s}" for s in scopes ]
        scope_string = " ".join(expanded_scopes)
        extended_scope_string=scope_string+extension
    return extended_scope_string

def ests_portal_login_flow(tenant, username, cookies):

    resp=requests.get(f"https://portal.azure.com:443/signin/index/@{tenant}?feature.argsubscriptions=true&feature.showservicehealthalerts=true&feature.prefetchtokens=true&feature.internalgraphapiversion=true&feature.selftoken=true&feature.globalresourcefilter=true&feature.msaljs=true&feature.fetchpolicyforrestypes=true&feature.testcrosscloudpuid=true&feature.useredirecthint=true&feature.usetenanthint=true&idpc=0")

    portalCookies=resp.cookies
    
    # Extract portal redirect URL
    portalRedirectUrl=string_between(resp.text, f"https://login.microsoftonline.com/{tenant}/oauth2/v2.0/authorize", "\")", include_start=True)

    # Begin auth code flow with portal-generated login flow
    resp=requests.get(portalRedirectUrl, cookies=cookies)

    # Get the login configurations for the client and tenant
    config_text=string_between(resp.text, "$Config=", ";\n//]]>")
    config=json.loads(config_text)

    # Extract the ctx token, flow token, and canary from the client configurations
    ctx=config["sCtx"]
    flow=config["sFT"]
    canary=config["apiCanary"]

    supported_mfa_methods=["PhoneAppNotification", "PhoneAppOTP", "OneWaySMS"]

    mfa_option_list=config["arrUserProofs"]

    options=[method["authMethodId"] for method in mfa_option_list ]

    option_list=list(map(str, range(len(options))))

    mfa_option_dict=dict(zip(option_list, options))

    text="Which MFA method would you like to pick?\n"
    for number, option in mfa_option_dict.items():
        if option not in supported_mfa_methods:
            text += f"{number}: {option}(not supported)\n"
        else:
            text += f"{number}: {option}\n"

    text += "\n"

    res=input(text)
    while True:
        if res not in option_list:
            print("Please pick a valid option...")
            res=input()
            continue
        if mfa_option_dict[res] not in supported_mfa_methods:
            print("Please pick a supported MFA option...")
            res=input()
            continue
        break


    mfaMethod=mfa_option_dict[res]

    # Start the authentication process.
    # Spoof the Edge user agent. Endpoint throws "Bad Reputation" error if python requests user agent is used.
    headers={
        "Accept": "application/json",
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.36 Edg/122.0.0.0" 
    }

    data={  
        "AuthMethodId":mfaMethod,
        "Method": "BeginAuth",
        "ctx": ctx,
        "flowToken": flow
    }

    resp=requests.post("https://login.microsoftonline.com/common/SAS/BeginAuth", json=data, headers=headers, allow_redirects=False)

    results=resp.json()
    if not results["Success"]:
        print(f"error while beginning mfa process. error: {results['Message']}. Exiting...")
        exit()

    # Based on the Auth method, complete MFA
    match mfaMethod:
        case "PhoneAppNotification":
            MFAChallenge=results["Entropy"]
            print(f"Please respond to the prompt on the authenticator app with the following number: {MFAChallenge}")

            while True:
                flow=results["FlowToken"]
                ctx=results["Ctx"]
                sessionId=results["SessionId"]

                query={
                    "authMethodId":mfaMethod
                }

                headers={
                    "X-Ms-Ctx": ctx,
                    "X-Ms-Flowtoken": flow,
                    "X-Ms-Sessionid": sessionId,
                    "Canary": canary
                }

                resp=requests.get( f"https://login.microsoftonline.com/common/SAS/EndAuth", headers=headers, params=query, allow_redirects=False )

                results=resp.json()
                if results["Success"]:
                    break
                sleep(3)

        case "PhoneAppOTP":

            poll_count=0

            while True:
                poll_count+=1
                flow=results["FlowToken"]
                ctx=results["Ctx"]
                sessionId=results["SessionId"]
                otp=input("Please enter an OTP from your phone authenticator app: ")
                headers={
                    "Canary": canary
                }

                data={
                    "Method": "EndAuth",
                    "Ctx": ctx,
                    "FlowToken": flow,
                    "AuthMethodId": mfaMethod,
                    "AdditionalAuthData": otp,
                    "pollCount": poll_count,
                    "SessionId": sessionId
                }

                resp=requests.post( f"https://login.microsoftonline.com/common/SAS/EndAuth", json=data, headers=headers, allow_redirects=False )

                results=resp.json()
                if results["Success"]:
                    break
                else:
                    print(f"Error: {results['Message']}")

        case "OneWaySMS":

            poll_count=0

            while True:
                poll_count+=1
                flow=results["FlowToken"]
                ctx=results["Ctx"]
                sessionId=results["SessionId"]

                otp=input("Please enter the code from the SMS: ")

                headers={
                    "Canary": canary
                }

                data={
                    "Method": "EndAuth",
                    "Ctx": ctx,
                    "FlowToken": flow,
                    "AuthMethodId": mfaMethod,
                    "AdditionalAuthData": otp,
                    "pollCount": poll_count,
                    "SessionId": sessionId
                }

                resp=requests.post( f"https://login.microsoftonline.com/common/SAS/EndAuth", json=data, headers=headers, allow_redirects=False )

                results=resp.json()
                if results["Success"]:
                    break
                else:
                    print(f"Error: {results['Message']}")
        case _:
            print( f"MFA Type {mfaMethod} not supported by azol. exiting")
            exit()

    flow=results["FlowToken"]
    ctx=results["Ctx"]
    sessionId=results["SessionId"]

    data={
        "request": ctx,
        "login": username,
        "flowToken": flow,
        "mfaAuthMethod": mfaMethod,
        "type":22
    }

    resp=requests.post( f"https://login.microsoftonline.com/common/SAS/ProcessAuth", data=data, allow_redirects=False)

    #look for kmsi interrupt. If no interrupt, eat exception and continue
    try:
        config_text=string_between(resp.text, "$Config=", ";\n//]]>")

        config=json.loads(config_text)
        
        flow=config["sFT"]
        canary=config["apiCanary"]
        ctx=config["sCtx"]
        pgid=config["pgid"]

        # If Keep Me Signed In(KMSI) interrupt, send "yes"
        if pgid=="KmsiInterrupt":
            data={
                "ctx": ctx,
                "flowToken": flow,
                "LoginOptions": 1
            }

            # kmsi = keep me signed in
            resp=requests.post("https://login.microsoftonline.com/kmsi", data=data, allow_redirects=False)

    except ValueError:
        pass

    secrets={}

    class SecretExtractor(HTMLParser):
        def handle_starttag(self, tag, attrs):

            name=None
            value=None
            if tag != "input": return
            for attr in attrs:
                if attr[0] == "name":
                    name=attr[1]
                if attr[0] == "value":
                    value=attr[1]
            if name is not None and value is not None:
                secrets[name]=value

    parser = SecretExtractor()
    parser.feed(resp.text)
    code=secrets["code"]
    if "state" in secrets.keys():
        state=urllib.parse.unquote(secrets["state"])
    if "id_token" in secrets.keys():
        id_token=secrets["id_token"]
    data={
        "state": state,
        "code": code,
        "id_token": id_token
    }

    headers={
        "Content-Type": "application/x-www-form-urlencoded"
    }

    resp=requests.post("https://portal.azure.com/signin/index/", data=data, allow_redirects=False, cookies=portalCookies, headers=headers)

    boot_params=string_between(resp.text, "MsPortalImpl.setBootstrapStateAndRedirect(", ");")
    portalbootstrapParams="[" + boot_params + "]"
    params=json.loads(portalbootstrapParams)

    portal_token_data=params[1]
    return portal_token_data

def ests_login_flow( tenant, client_id, ests, ests_persistent, username, scope_string, redirect_url ):
    # if ESTSAUTH and ESTSAUTHPERSISTENT are both present, we can only use one.
    # request will fail if both are used. preference should be ESTSAUTHPERSISTENT
    if ests_persistent is not None:
        cookies={
            "ESTSAUTHPERSISTENT": ests_persistent
        }
    elif ests_persistent is None and ests is not None:
        cookies={
            "ESTSAUTH": ests
        }
    else:
        raise Exception()
    
    if client_id==FOCIClients.AzurePortal:
        portalTokens=ests_portal_login_flow(tenant, username, cookies)
        code=portalTokens["spaAuthCode"]
    else:
        code_challenge=''.join(random.choices(string.ascii_letters, k=43))
        queryParams= {
            "redirect_uri": redirect_url,
            "response_type": "code",
            "scope": scope_string,
            "response_mode": "form_post",
            "client_id": client_id,
            "code_challenge":code_challenge,
            "code_challenge_method":"plain"
        }
        resp=requests.get(f"https://login.microsoftonline.com/{tenant}/oauth2/v2.0/authorize", params=queryParams, allow_redirects=False, cookies=cookies)


        # Get the login configurations for the client and tenant
        config_text=string_between(resp.text, "$Config=", ";\n//]]>")
        config=json.loads(config_text)

        # Extract the ctx token, flow token, and canary from the client configurations
        ctx=config["sCtx"]
        flow=config["sFT"]
        canary=config["apiCanary"]

        supported_mfa_methods=["PhoneAppNotification", "PhoneAppOTP", "OneWaySMS"]

        mfa_option_list=config["arrUserProofs"]

        options=[method["authMethodId"] for method in mfa_option_list ]

        option_list=list(map(str, range(len(options))))

        mfa_option_dict=dict(zip(option_list, options))

        text="Which MFA method would you like to pick?\n"
        for number, option in mfa_option_dict.items():
            if option not in supported_mfa_methods:
                text += f"{number}: {option}(not supported)\n"
            else:
                text += f"{number}: {option}\n"

        text += "\n"

        res=input(text)
        while True:
            if res not in option_list:
                print("Please pick a valid option...")
                res=input()
                continue
            if mfa_option_dict[res] not in supported_mfa_methods:
                print("Please pick a supported MFA option...")
                res=input()
                continue
            break


        mfaMethod=mfa_option_dict[res]

        # Start the authentication process.
        # Spoof the Edge user agent. Endpoint throws "Bad Reputation" error if python requests user agent is used.
        headers={
            "Accept": "application/json",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.36 Edg/122.0.0.0" 
        }

        data={  
            "AuthMethodId":mfaMethod,
            "Method": "BeginAuth",
            "ctx": ctx,
            "flowToken": flow
        }

        resp=requests.post("https://login.microsoftonline.com/common/SAS/BeginAuth", json=data, headers=headers, allow_redirects=False)

        results=resp.json()
        if not results["Success"]:
            print(f"error while beginning mfa process. error: {results['Message']}. Exiting...")
            exit()

            # Based on the Auth method, complete MFA
        match mfaMethod:
            case "PhoneAppNotification":
                MFAChallenge=results["Entropy"]
                print(f"Please respond to the prompt on the authenticator app with the following number: {MFAChallenge}")

                while True:
                    flow=results["FlowToken"]
                    ctx=results["Ctx"]
                    sessionId=results["SessionId"]

                    query={
                        "authMethodId":mfaMethod
                    }

                    headers={
                        "X-Ms-Ctx": ctx,
                        "X-Ms-Flowtoken": flow,
                        "X-Ms-Sessionid": sessionId,
                        "Canary": canary
                    }

                    resp=requests.get( f"https://login.microsoftonline.com/common/SAS/EndAuth", headers=headers, params=query, allow_redirects=False )

                    results=resp.json()
                    if results["Success"]:
                        break
                    sleep(3)

            case "PhoneAppOTP":

                poll_count=0

                while True:
                    poll_count+=1
                    flow=results["FlowToken"]
                    ctx=results["Ctx"]
                    sessionId=results["SessionId"]
                    otp=input("Please enter an OTP from your phone authenticator app: ")
                    headers={
                        "Canary": canary
                    }

                    data={
                        "Method": "EndAuth",
                        "Ctx": ctx,
                        "FlowToken": flow,
                        "AuthMethodId": mfaMethod,
                        "AdditionalAuthData": otp,
                        "pollCount": poll_count,
                        "SessionId": sessionId
                    }

                    resp=requests.post( f"https://login.microsoftonline.com/common/SAS/EndAuth", json=data, headers=headers, allow_redirects=False )

                    results=resp.json()
                    if results["Success"]:
                        break
                    else:
                        print(f"Error: {results['Message']}")

            case "OneWaySMS":

                poll_count=0

                while True:
                    poll_count+=1
                    flow=results["FlowToken"]
                    ctx=results["Ctx"]
                    sessionId=results["SessionId"]

                    otp=input("Please enter the code from the SMS: ")

                    headers={
                        "Canary": canary
                    }

                    data={
                        "Method": "EndAuth",
                        "Ctx": ctx,
                        "FlowToken": flow,
                        "AuthMethodId": mfaMethod,
                        "AdditionalAuthData": otp,
                        "pollCount": poll_count,
                        "SessionId": sessionId
                    }

                    resp=requests.post( f"https://login.microsoftonline.com/common/SAS/EndAuth", json=data, headers=headers, allow_redirects=False )

                    results=resp.json()
                    if results["Success"]:
                        break
                    else:
                        print(f"Error: {results['Message']}")
            case _:
                print( f"MFA Type {mfaMethod} not supported by azol. exiting")
                exit()

        flow=results["FlowToken"]
        ctx=results["Ctx"]
        sessionId=results["SessionId"]

        data={
            "request": ctx,
            "login": username,
            "flowToken": flow,
            "mfaAuthMethod": mfaMethod,
            "type":22
        }

        resp=requests.post( f"https://login.microsoftonline.com/common/SAS/ProcessAuth", data=data, allow_redirects=False)
        

        #look for kmsi interrupt. If no interrupt, eat exception and continue
        try:
            config_text=string_between(resp.text, "$Config=", ";\n//]]>")
                
            config=json.loads(config_text)
            
            flow=config["sFT"]
            canary=config["apiCanary"]
            ctx=config["sCtx"]
            pgid=config["pgid"]

            # If Keep Me Signed In(KMSI) interrupt, send "yes"
            if pgid=="KmsiInterrupt":
                data={
                    "ctx": ctx,
                    "flowToken": flow,
                    "LoginOptions": 1
                }

                # kmsi = keep me signed in
                resp=requests.post("https://login.microsoftonline.com/kmsi", data=data, allow_redirects=False)

        except ValueError:
            pass

        secrets={}

        class SecretExtractor(HTMLParser):
            def handle_starttag(self, tag, attrs):

                name=None
                value=None
                if tag != "input": return
                for attr in attrs:
                    if attr[0] == "name":
                        name=attr[1]
                    if attr[0] == "value":
                        value=attr[1]
                if name is not None and value is not None:
                    secrets[name]=value

        parser = SecretExtractor()
        parser.feed(resp.text)
        code=secrets["code"]
        state=None
        id_token=None
        if "state" in secrets.keys():
            state=urllib.parse.unquote(secrets["state"])
        if "id_token" in secrets.keys():
            id_token=secrets["id_token"]
        data={
                "code": code
            }
        if state is not None:
            data["state"]=state
        if id_token is not None:
            data["id_token"]=id_token

        headers={
            "Content-Type": "application/x-www-form-urlencoded"
        }

    if client_id==FOCIClients.AzurePortal:
        data={
            "client_id": client_id,
            "code": code,
            "scope": scope_string,
            "grant_type":"authorization_code"
        }
    
    else:
        data={
            "client_id": client_id,
            "code": code,
            "scope": scope_string,
            "grant_type":"authorization_code",
            "redirect_uri": redirect_url,
            "code_verifier": code_challenge
        }

    headers={
        "origin": "http://localhost"
    }

    resp=requests.post(f"https://login.microsoftonline.com/{tenant}/oauth2/v2.0/token", data=data, allow_redirects=False, headers=headers)

    json_response=resp.json()
    new_token_data = {
        "access_token": json_response["access_token"],
        "refresh_token": json_response["refresh_token"],
        "flow_type": "delegated"
    }
    return new_token_data

def get_portal_tokens(tenant, username):
    ests=None
    ests_persistent=None
    # Azure portal has a weird auth code flow. Perform extra steps
    # Azure Portal step 1: get portal-generated redirect uri with nonce
    resp=requests.get(f"https://portal.azure.com:443/signin/index/@{tenant}?feature.argsubscriptions=true&feature.showservicehealthalerts=true&feature.prefetchtokens=true&feature.internalgraphapiversion=true&feature.selftoken=true&feature.globalresourcefilter=true&feature.msaljs=true&feature.fetchpolicyforrestypes=true&feature.testcrosscloudpuid=true&feature.useredirecthint=true&feature.usetenanthint=true&idpc=0")

    # Save Portal cookies for later - required when finishing login flow
    portalCookies=resp.cookies

    # Extract portal redirect URL
    portalRedirectUrl=string_between(resp.text, f"https://login.microsoftonline.com/{tenant}/oauth2/v2.0/authorize", "\")", include_start=True)

    # Begin auth code flow with portal-generated login flow
    resp=requests.get(portalRedirectUrl)

    # Get the login configurations for the client and tenant
    config_text=string_between(resp.text, "$Config=", ";\n//]]>")
    config=json.loads(config_text)

    # Extract the ctx token, flow token, and canary from the client configurations
    ctx=config["sCtx"]
    flow=config["sFT"]
    canary=config["apiCanary"]

    headers={
        "Canary": canary
    }

    data={
        "username": username,
        "isOtherIdpSupported": "true",
        "checkPhones": "false",
        "isRemoteNGCSupported": "false",
        "isCookieBannerShown": "false",
        "isFidoSupported": "false",
        "originalRequest": ctx,
        "forceotclogin": "false",
        "isExternalFederationDisallowed": "false",
        "isRemoteConnectSupported": "false",
        "federationFlags": "0",
        "isSignup": "false",
        "isAccessPassSupported": "true",
        "flowToken": flow,
        "Country": "NO"
    }

    resp=requests.post(f"https://login.microsoftonline.com/{tenant}/GetCredentialType", json=data, headers=headers, allow_redirects=False)

    userLoginParams=resp.json()
    flow=userLoginParams["FlowToken"]
    canary=userLoginParams["apiCanary"]
    hasPassword=userLoginParams["Credentials"]["HasPassword"]

    while True:
        if hasPassword:
            password = getpass.getpass('Password:')

        data={
            "login": username,
            "loginfmt": username,
            "passwd": password,
            "ctx": ctx,
            "flowToken": flow,
            "LoginOptions": 3
        }

        resp=requests.post(f"https://login.microsoftonline.com/{tenant}/login", data=data, allow_redirects=False)
        
        # if an exception occurs, there is probably no MFA requirement. code is returned in a form_post
        try:
            config_text=string_between(resp.text, "$Config=", ";\n//]]>")
            config=json.loads(config_text)
            if "arrUserProofs" not in config.keys():
                require_mfa=False
            if "sErrorCode" in config.keys():
                if config["sErrorCode"] == "50126":
                    print("Wrong password. Try again.")
            else:
                break
        except:
            require_mfa=False
            break

    if "arrUserProofs" in config.keys():
        require_mfa=True

    if require_mfa:
        flow=config["sFT"]
        canary=config["apiCanary"]
        ctx=config["sCtx"]

        supported_mfa_methods=["PhoneAppNotification", "PhoneAppOTP", "OneWaySMS"]

        mfa_option_list=config["arrUserProofs"]

        options=[method["authMethodId"] for method in mfa_option_list ]

        option_list=list(map(str, range(len(options))))

        mfa_option_dict=dict(zip(option_list, options))

        text="Which MFA method would you like to pick?\n"
        for number, option in mfa_option_dict.items():
            if option not in supported_mfa_methods:
                text += f"{number}: {option}(not supported)\n"
            else:
                text += f"{number}: {option}\n"

        text += "\n"

        res=input(text)
        while True:
            if res not in option_list:
                print("Please pick a valid option...")
                res=input()
                continue
            if mfa_option_dict[res] not in supported_mfa_methods:
                print("Please pick a supported MFA option...")
                res=input()
                continue
            break


        mfaMethod=mfa_option_dict[res]

        # Start the authentication process.
        # Spoof the Edge user agent. Endpoint throws "Bad Reputation" error if python requests user agent is used.
        headers={
            "Accept": "application/json",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.36 Edg/122.0.0.0" 
        }

        data={  
            "AuthMethodId":mfaMethod,
            "Method": "BeginAuth",
            "ctx": ctx,
            "flowToken": flow
        }

        resp=requests.post("https://login.microsoftonline.com/common/SAS/BeginAuth", json=data, headers=headers, allow_redirects=False)

        results=resp.json()
        if not results["Success"]:
            print(f"error while beginning mfa process. error: {results['Message']}. Exiting...")
            exit()

         # Based on the Auth method, complete MFA
        match mfaMethod:
            case "PhoneAppNotification":
                MFAChallenge=results["Entropy"]
                print(f"Please respond to the prompt on the authenticator app with the following number: {MFAChallenge}")

                while True:
                    flow=results["FlowToken"]
                    ctx=results["Ctx"]
                    sessionId=results["SessionId"]

                    query={
                        "authMethodId":mfaMethod
                    }

                    headers={
                        "X-Ms-Ctx": ctx,
                        "X-Ms-Flowtoken": flow,
                        "X-Ms-Sessionid": sessionId,
                        "Canary": canary
                    }

                    resp=requests.get( f"https://login.microsoftonline.com/common/SAS/EndAuth", headers=headers, params=query, allow_redirects=False )

                    results=resp.json()
                    if results["Success"]:
                        break
                    sleep(3)

            case "PhoneAppOTP":

                poll_count=0

                while True:
                    poll_count+=1
                    flow=results["FlowToken"]
                    ctx=results["Ctx"]
                    sessionId=results["SessionId"]
                    otp=input("Please enter an OTP from your phone authenticator app: ")
                    headers={
                        "Canary": canary
                    }

                    data={
                        "Method": "EndAuth",
                        "Ctx": ctx,
                        "FlowToken": flow,
                        "AuthMethodId": mfaMethod,
                        "AdditionalAuthData": otp,
                        "pollCount": poll_count,
                        "SessionId": sessionId
                    }

                    resp=requests.post( f"https://login.microsoftonline.com/common/SAS/EndAuth", json=data, headers=headers, allow_redirects=False )

                    results=resp.json()
                    if results["Success"]:
                        break
                    else:
                        print(f"Error: {results['Message']}")

            case "OneWaySMS":

                poll_count=0

                while True:
                    poll_count+=1
                    flow=results["FlowToken"]
                    ctx=results["Ctx"]
                    sessionId=results["SessionId"]

                    otp=input("Please enter the code from the SMS: ")

                    headers={
                        "Canary": canary
                    }

                    data={
                        "Method": "EndAuth",
                        "Ctx": ctx,
                        "FlowToken": flow,
                        "AuthMethodId": mfaMethod,
                        "AdditionalAuthData": otp,
                        "pollCount": poll_count,
                        "SessionId": sessionId
                    }

                    resp=requests.post( f"https://login.microsoftonline.com/common/SAS/EndAuth", json=data, headers=headers, allow_redirects=False )

                    results=resp.json()
                    if results["Success"]:
                        break
                    else:
                        print(f"Error: {results['Message']}")
            case _:
                print( f"MFA Type {mfaMethod} not supported by azol. exiting")
                exit()

        flow=results["FlowToken"]
        ctx=results["Ctx"]
        sessionId=results["SessionId"]

        data={
            "request": ctx,
            "login": username,
            "flowToken": flow,
            "mfaAuthMethod": mfaMethod,
            "type":22
        }

        resp=requests.post( f"https://login.microsoftonline.com/common/SAS/ProcessAuth", data=data, allow_redirects=False)
        

    #look for kmsi interrupt. If no interrupt, eat exception and continue
    try:
        config_text=string_between(resp.text, "$Config=", ";\n//]]>")

        if "ESTSAUTH" in resp.cookies.keys():
            ests= resp.cookies.get("ESTSAUTH")

            
        config=json.loads(config_text)
        
        flow=config["sFT"]
        canary=config["apiCanary"]
        ctx=config["sCtx"]
        pgid=config["pgid"]

        # If Keep Me Signed In(KMSI) interrupt, send "yes"
        if pgid=="KmsiInterrupt":
            data={
                "ctx": ctx,
                "flowToken": flow,
                "LoginOptions": 1
            }

            # kmsi = keep me signed in
            resp=requests.post("https://login.microsoftonline.com/kmsi", data=data, allow_redirects=False)

            if "ESTSAUTHPERSISTENT" in resp.cookies.keys():
                estspersistent= resp.cookies.get("ESTSAUTHPERSISTENT")

    except ValueError:
        pass

    secrets={}

    class SecretExtractor(HTMLParser):
        def handle_starttag(self, tag, attrs):

            name=None
            value=None
            if tag != "input": return
            for attr in attrs:
                if attr[0] == "name":
                    name=attr[1]
                if attr[0] == "value":
                    value=attr[1]
            if name is not None and value is not None:
                secrets[name]=value

    parser = SecretExtractor()
    parser.feed(resp.text)
    code=secrets["code"]
    if "state" in secrets.keys():
        state=urllib.parse.unquote(secrets["state"])
    if "id_token" in secrets.keys():
        id_token=secrets["id_token"]
    data={
        "state": state,
        "code": code,
        "id_token": id_token
    }

    headers={
        "Content-Type": "application/x-www-form-urlencoded"
    }

    resp=requests.post("https://portal.azure.com/signin/index/", data=data, allow_redirects=False, cookies=portalCookies, headers=headers)

    boot_params=string_between(resp.text, "MsPortalImpl.setBootstrapStateAndRedirect(", ");")
    portalbootstrapParams="[" + boot_params + "]"
    params=json.loads(portalbootstrapParams)

    token_data=params[1]
    token_data["ESTSAUTH"]=ests
    token_data["ESTSAUTHPERSISTENT"]=estspersistent
    return token_data

def auth_code_flow(tenant, username, client_id, redirect_url, scope_string):
    ests=None
    ests_persistent=None
    if client_id == FOCIClients.AzurePortal:
        portalTokens=get_portal_tokens(tenant, username)
        code=portalTokens["spaAuthCode"]

    else:
        code_challenge=''.join(random.choices(string.ascii_letters, k=43))
        queryParams= {
            "redirect_uri": redirect_url,
            "response_type": "code",
            "scope": scope_string,
            "response_mode": "form_post",
            "client_id": client_id,
            "code_challenge":code_challenge,
            "code_challenge_method":"plain"
        }
        resp=requests.get(f"https://login.microsoftonline.com/{tenant}/oauth2/v2.0/authorize", params=queryParams, allow_redirects=False)

        # Get the login configurations for the client and tenant
        config_text=string_between(resp.text, "$Config=", ";\n//]]>")
        config=json.loads(config_text)

        # Extract the ctx token, flow token, and canary from the client configurations
        ctx=config["sCtx"]
        flow=config["sFT"]
        canary=config["apiCanary"]

        headers={
            "Canary": canary
        }

        data={
            "username": username,
            "isOtherIdpSupported": "true",
            "checkPhones": "false",
            "isRemoteNGCSupported": "false",
            "isCookieBannerShown": "false",
            "isFidoSupported": "false",
            "originalRequest": ctx,
            "forceotclogin": "false",
            "isExternalFederationDisallowed": "false",
            "isRemoteConnectSupported": "false",
            "federationFlags": "0",
            "isSignup": "false",
            "isAccessPassSupported": "true",
            "flowToken": flow,
            "Country": "NO"
        }

        resp=requests.post(f"https://login.microsoftonline.com/{tenant}/GetCredentialType", json=data, headers=headers, allow_redirects=False)

        userLoginParams=resp.json()
        flow=userLoginParams["FlowToken"]
        canary=userLoginParams["apiCanary"]
        hasPassword=userLoginParams["Credentials"]["HasPassword"]

        while True:
            if hasPassword:
                password = getpass.getpass('Password:')

            data={
                "login": username,
                "loginfmt": username,
                "passwd": password,
                "ctx": ctx,
                "flowToken": flow,
                "LoginOptions": 3
            }

            resp=requests.post(f"https://login.microsoftonline.com/{tenant}/login", data=data, allow_redirects=False)
            
            # if an exception occurs, there is probably no MFA requirement. code is returned in a form_post
            try:
                config_text=string_between(resp.text, "$Config=", ";\n//]]>")
                config=json.loads(config_text)
                if "arrUserProofs" not in config.keys():
                    require_mfa=False
                if "sErrorCode" in config.keys():
                    if config["sErrorCode"] == "50126":
                        print("Wrong password. Try again.")
                else:
                    break
            except:
                require_mfa=False
                break

        if "arrUserProofs" in config.keys():
            require_mfa=True

        if require_mfa:
            flow=config["sFT"]
            canary=config["apiCanary"]
            ctx=config["sCtx"]

            supported_mfa_methods=["PhoneAppNotification", "PhoneAppOTP", "OneWaySMS"]

            mfa_option_list=config["arrUserProofs"]

            options=[method["authMethodId"] for method in mfa_option_list ]

            option_list=list(map(str, range(len(options))))

            mfa_option_dict=dict(zip(option_list, options))

            text="Which MFA method would you like to pick?\n"
            for number, option in mfa_option_dict.items():
                if option not in supported_mfa_methods:
                    text += f"{number}: {option}(not supported)\n"
                else:
                    text += f"{number}: {option}\n"

            text += "\n"

            res=input(text)
            while True:
                if res not in option_list:
                    print("Please pick a valid option...")
                    res=input()
                    continue
                if mfa_option_dict[res] not in supported_mfa_methods:
                    print("Please pick a supported MFA option...")
                    res=input()
                    continue
                break


            mfaMethod=mfa_option_dict[res]

            # Start the authentication process.
            # Spoof the Edge user agent. Endpoint throws "Bad Reputation" error if python requests user agent is used.
            headers={
                "Accept": "application/json",
                "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.36 Edg/122.0.0.0" 
            }

            data={  
                "AuthMethodId":mfaMethod,
                "Method": "BeginAuth",
                "ctx": ctx,
                "flowToken": flow
            }

            resp=requests.post("https://login.microsoftonline.com/common/SAS/BeginAuth", json=data, headers=headers, allow_redirects=False)

            results=resp.json()
            if not results["Success"]:
                print(f"error while beginning mfa process. error: {results['Message']}. Exiting...")
                exit()

            # Based on the Auth method, complete MFA
            match mfaMethod:
                case "PhoneAppNotification":
                    MFAChallenge=results["Entropy"]
                    print(f"Please respond to the prompt on the authenticator app with the following number: {MFAChallenge}")

                    while True:
                        flow=results["FlowToken"]
                        ctx=results["Ctx"]
                        sessionId=results["SessionId"]

                        query={
                            "authMethodId":mfaMethod
                        }

                        headers={
                            "X-Ms-Ctx": ctx,
                            "X-Ms-Flowtoken": flow,
                            "X-Ms-Sessionid": sessionId,
                            "Canary": canary
                        }

                        resp=requests.get( f"https://login.microsoftonline.com/common/SAS/EndAuth", headers=headers, params=query, allow_redirects=False )

                        results=resp.json()
                        if results["Success"]:
                            break
                        sleep(3)

                case "PhoneAppOTP":

                    poll_count=0

                    while True:
                        poll_count+=1
                        flow=results["FlowToken"]
                        ctx=results["Ctx"]
                        sessionId=results["SessionId"]
                        otp=input("Please enter an OTP from your phone authenticator app: ")
                        headers={
                            "Canary": canary
                        }

                        data={
                            "Method": "EndAuth",
                            "Ctx": ctx,
                            "FlowToken": flow,
                            "AuthMethodId": mfaMethod,
                            "AdditionalAuthData": otp,
                            "pollCount": poll_count,
                            "SessionId": sessionId
                        }

                        resp=requests.post( f"https://login.microsoftonline.com/common/SAS/EndAuth", json=data, headers=headers, allow_redirects=False )

                        results=resp.json()
                        if results["Success"]:
                            break
                        else:
                            print(f"Error: {results['Message']}")

                case "OneWaySMS":

                    poll_count=0

                    while True:
                        poll_count+=1
                        flow=results["FlowToken"]
                        ctx=results["Ctx"]
                        sessionId=results["SessionId"]

                        otp=input("Please enter the code from the SMS: ")

                        headers={
                            "Canary": canary
                        }

                        data={
                            "Method": "EndAuth",
                            "Ctx": ctx,
                            "FlowToken": flow,
                            "AuthMethodId": mfaMethod,
                            "AdditionalAuthData": otp,
                            "pollCount": poll_count,
                            "SessionId": sessionId
                        }

                        resp=requests.post( f"https://login.microsoftonline.com/common/SAS/EndAuth", json=data, headers=headers, allow_redirects=False )

                        results=resp.json()
                        if results["Success"]:
                            break
                        else:
                            print(f"Error: {results['Message']}")
                case _:
                    print( f"MFA Type {mfaMethod} not supported by azol. exiting")
                    exit()

            flow=results["FlowToken"]
            ctx=results["Ctx"]
            sessionId=results["SessionId"]

            data={
                "request": ctx,
                "login": username,
                "flowToken": flow,
                "mfaAuthMethod": mfaMethod,
                "type":22
            }

            resp=requests.post( f"https://login.microsoftonline.com/common/SAS/ProcessAuth", data=data, allow_redirects=False)
            
            if "ESTSAUTH" in resp.cookies.keys():
                ests= resp.cookies.get("ESTSAUTH")
            

        #look for kmsi interrupt. If no interrupt, eat exception and continue
        try:
            config_text=string_between(resp.text, "$Config=", ";\n//]]>")
            config=json.loads(config_text)
            flow=config["sFT"]
            canary=config["apiCanary"]
            ctx=config["sCtx"]
            pgid=config["pgid"]

            # If Keep Me Signed In(KMSI) interrupt, send "yes"
            if pgid=="KmsiInterrupt":
                data={
                    "ctx": ctx,
                    "flowToken": flow,
                    "LoginOptions": 1
                }

                # kmsi = keep me signed in
                resp=requests.post("https://login.microsoftonline.com/kmsi", data=data, allow_redirects=False)
                if "ESTSAUTHPERSISTENT" in resp.cookies.keys():
                    ests_persistent= resp.cookies.get("ESTSAUTHPERSISTENT")
        except ValueError:
            pass

        secrets={}

        class SecretExtractor(HTMLParser):
            def handle_starttag(self, tag, attrs):

                name=None
                value=None
                if tag != "input": return
                for attr in attrs:
                    if attr[0] == "name":
                        name=attr[1]
                    if attr[0] == "value":
                        value=attr[1]
                if name is not None and value is not None:
                    secrets[name]=value

        parser = SecretExtractor()
        parser.feed(resp.text)
        code=secrets["code"]
        if "state" in secrets.keys():
            state=urllib.parse.unquote(secrets["state"])
        if "id_token" in secrets.keys():
            id_token=secrets["id_token"]

    if client_id==FOCIClients.AzurePortal:
        data={
            "client_id": client_id,
            "code": code,
            "scope": scope_string,
            "grant_type":"authorization_code"
        }
        ests=portalTokens["ESTSAUTH"]
        ests_persistent=portalTokens["ESTSAUTHPERSISTENT"]
    
    else:
        data={
            "client_id": client_id,
            "code": code,
            "scope": scope_string,
            "grant_type":"authorization_code",
            "redirect_uri": redirect_url,
            "code_verifier": code_challenge
        }
    
    headers={
        "origin": "http://localhost"
    }

    resp=requests.post(f"https://login.microsoftonline.com/{tenant}/oauth2/v2.0/token", data=data, allow_redirects=False, headers=headers)

    json_response=resp.json()
    new_token_data = {
        "access_token": json_response["access_token"],
        "refresh_token": json_response["refresh_token"],
        "flow_type": "delegated",
        "ests": ests,
        "estspersistent": ests_persistent
    }
    return new_token_data


"""Module containing all http clients"""
from azol.clients.arm_client import ArmClient
from azol.clients.graph_client import GraphClient
from azol.clients.key_vault_client import KeyVaultClient
from azol.clients.oauth_http_client import OAuthHTTPClient
from azol.clients.datafactory_client import DataFactoryClient

__all__ = [
    "ArmClient",
    "GraphClient",
    "KeyVaultClient",
    "OAuthHTTPClient",
    "DataFactoryClient"
]



"""Module containing useful resources that are used, for example, for data annotation in azol"""
from .graph_delegated_permissions import delegated_permissions
from .graph_permissions_map import graph_permission_map
from .rbac_roles import rbac_roles



rbac_roles = {
    "00482a5a-887f-4fb3-b363-3b7fe8e74483": "Key Vault Administrator",
    "00493d72-78f6-4148-b6c5-d3ce8e4799dd": "Azure Arc Enabled Kubernetes Cluster User Role",
    "00c29273-979b-4161-815c-10b084fb9324": "Backup Operator",
    "0105a6b0-4bb9-43d2-982a-12806f9faddb": "Azure Center for SAP solutions Service role for management",
    "028f4ed7-e2a9-465e-a8f4-9c0ffdfdc027": "Data Box Reader",
    "02ca0879-e8e4-47a5-a61e-5c618b76e64a": "Device Update Administrator",
    "0378884a-3af5-44ab-8323-f5b22f9f3c98": "Device Update Content Administrator",
    "03a6d094-3444-4b3d-88af-7477090a9e5e": "Intelligent Systems Account Contributor",
    "04165923-9d83-45d5-8227-78b77b0a687e": "SignalR AccessKey Reader",
    "05352d14-a920-4328-a0de-4cbe7430e26b": "Azure Center for SAP solutions reader",
    "054126f8-9a2b-4f1c-a9ad-eca461f08466": "Media Services Account Administrator",
    "056cd41c-7e88-42e1-933e-88ba6a50c9c3": "SQL Security Manager",
    "05b7651b-dc44-475e-b74d-df3db49fae0f": "Project Babylon Data Source Administrator",
    "082f0a83-3be5-4ba1-904c-961cca79b387": "Desktop Virtualization Contributor",
    "088ab73d-1256-47ae-bea9-9de8e7131f31": "Guest Configuration Resource Contributor",
    "08954f03-6346-4c2e-81c0-ec3a5cfae23b": "Application Insights Snapshot Debugger",
    "090c5cfd-751d-490a-894a-3ce6f1109419": "Azure Service Bus Data Owner",
    "0a5ae4ab-0d65-4eeb-be61-29fc9b54394b": "Cognitive Services Custom Vision Trainer",
    "0a9a7e1f-b9d0-4cc4-a60d-0319b160aaa3": "Storage Table Data Contributor",
    "0ab0b1a8-8aac-4efd-b8c2-3ee1fb270be8": "Azure Kubernetes Service Cluster Admin Role",
    "0ab34830-df19-4f8c-b84e-aa85b8afa6e8": "Azure Front Door Domain Contributor",
    "0b555d9b-b4a7-4f43-b330-627f0e5be8f0": "Security Detonation Chamber Submitter",
    "0c34c906-8d99-4cb7-8bb7-33f5b0a1a799": "API Management Workspace Contributor",
    "0c867c2a-1d8c-454a-a3db-ab2ea1bdc8bb": "Storage File Data SMB Share Contributor",
    "0c8b84dc-067c-4039-9615-fa1a4b77c726": "PlayFab Contributor",
    "0db238c4-885e-4c4f-a933-aa2cef684fca": "Azure Front Door Secret Reader",
    "0e5f05e5-9ab9-446b-b98d-1e2157c94125": "Quota Request Operator",
    "0e75ca1e-0464-4b4d-8b93-68208a576181": "Cognitive Services Speech Contributor",
    "0f2ebee7-ffd4-4fc0-b3b7-664099fdad5d": "DNS Resolver Contributor",
    "0f37683f-2463-46b6-9ce7-9b788b988ba2": "App Compliance Automation Administrator",
    "0f99d363-226e-4dca-9920-b807cf8e1a5f": "Azure Front Door Domain Reader",
    "0fa44ee9-7a7d-466b-9bb2-2bf446b1204d": "Desktop Virtualization Workspace Reader",
    "10745317-c249-44a1-a5ce-3a4353c0bbd8": "Device Provisioning Service Data Reader",
    "12338af0-0e69-4776-bea7-57ae8d297424": "Key Vault Crypto User",
    "12cf5a90-567b-43ae-8102-96cf46c7d9b4": "Web PubSub Service Owner",
    "1407120a-92aa-4202-b7e9-c0e197c71c8f": "Search Index Data Reader",
    "14b46e9e-c2b7-41b4-b07b-48a6ebf60603": "Key Vault Crypto Officer",
    "150f5e0c-0603-4f03-8c7f-cf70034c4e90": "Data Purger",
    "15e0f5a1-3450-4248-8e25-e2afe88a9e85": "Test Base Reader",
    "17d1049b-9a84-46fb-8f53-869881c3d3ab": "Storage Account Contributor",
    "1823dd4f-9b8c-4ab6-ab4e-7397a3684615": "AzureML Registry User",
    "18500a29-7fe2-46b2-a342-b16a415e101d": "Managed HSM contributor",
    "188a0f2f-5c9e-469b-ae67-2aa5ce574b94": "Scheduler Job Collections Contributor",
    "189207d4-bb67-4208-a635-b06afe8b2c57": "SqlDb Migration Role",
    "18ab4d3d-a1bf-4477-8ad9-8359bc988f69": "Azure Kubernetes Fleet Manager RBAC Cluster Admin",
    "18d7d88d-d35e-4fb5-a5c3-7773c20a72d9": "User Access Administrator",
    "18e40d4e-8d2e-438d-97e1-9528336e149c": "Deployment Environments User",
    "18e81cdc-4e98-4e29-a639-e7d10c5a6226": "Cognitive Services LUIS Reader",
    "18ed5180-3e48-46fd-8541-4ea054d57064": "Azure Kubernetes Service Policy Add-on Deployment",
    "19e7f393-937e-4f77-808e-94535e297925": "Storage Queue Data Reader",
    "1afdec4b-e479-420e-99e7-f82237c7c5e6": "Azure Kubernetes Service Cluster Monitoring User",
    "1c0163c0-47e6-4577-8991-ea5c82e286e4": "Virtual Machine Administrator Login",
    "1c9b6475-caf0-4164-b5a1-2142a7116f4b": "Template Spec Contributor",
    "1d18fff3-a72a-46b5-b4a9-0b38a3cd7e63": "Desktop Virtualization User",
    "1d335eef-eee1-47fe-a9e0-53214eba8872": "SqlMI Migration Role",
    "1e241071-0855-49ea-94dc-649edcd759de": "EventGrid Contributor",
    "1ec5b3c1-b17e-4e25-8312-2acb3c3c5abf": "Stream Analytics Query Tester",
    "1ef6a3be-d0ac-425d-8c01-acb62866290b": "Compute Gallery Sharing Admin",
    "1f135831-5bbe-4924-9016-264044c00788": "Windows365NetworkInterfaceContributor",
    "200bba9e-f0c8-430f-892b-6f0794863803": "Purview role 2 (Deprecated)",
    "21090545-7ca7-4776-b22c-e363652d74d2": "Key Vault Reader",
    "21efdde3-836f-432b-bf3d-3e8e734d4b2b": "Desktop Virtualization Workspace Contributor",
    "22926164-76b3-42b3-bc55-97df8dab3e41": "Grafana Admin",
    "230815da-be43-4aae-9cb4-875f7bd000aa": "Cosmos DB Operator",
    "2414bbcf-6497-4faf-8c65-045460748405": "EventGrid EventSubscription Reader",
    "25fbc0a9-bd7c-42a3-aa1a-3b75d497ee68": "Cognitive Services Contributor",
    "26baccc8-eea7-41f1-98f4-1762cc7f685d": "Azure Relay Sender",
    "26e0b698-aa6d-4085-9386-aadae190014d": "Azure Relay Listener",
    "2787bf04-f1f5-4bfe-8383-c8a24483ee38": "Azure Relay Owner",
    "27f8b550-c507-4db9-86f2-f4b8e816d59d": "Autonomous Development Platform Data Owner (Preview)",
    "28241645-39f8-410b-ad48-87863e2951d5": "Security Detonation Chamber Reader",
    "2837e146-70d7-4cfd-ad55-7efa6464f958": "Code Signing Certificate Profile Signer",
    "2a2b9908-6ea1-4ae2-8e65-a410df84e7d1": "Storage Blob Data Reader",
    "2a5c394f-5eb7-4d4f-9c8e-e8eae39faebc": "Lab Services Reader",
    "2ad6aaab-ead9-4eaa-8ac5-da422f562408": "Desktop Virtualization Session Host Operator",
    "2b629674-e913-4c01-ae53-ef4638d8f975": "Azure Event Hubs Data Sender",
    "2c56ea50-c6b3-40a6-83c0-9d98858bc7d2": "Schema Registry Reader (Preview)",
    "2cc479cb-7b4d-49a8-b449-8c00fd0f0a4b": "Web Plan Contributor",
    "30b27cfc-9c84-438e-b0ce-70e35255df80": "Azure Kubernetes Fleet Manager RBAC Reader",
    "312a565d-c81f-4fd8-895a-4e21e48d571c": "API Management Service Contributor",
    "31a002a1-acaf-453e-8a5b-297c9ca1ea24": "Blockchain Member Node Access (Preview)",
    "331c37c6-af14-46d9-b9f4-e1909e1b95a0": "DevCenter Project Admin",
    "3498e952-d568-435e-9b2c-8d77e338d7f7": "Azure Kubernetes Service RBAC Admin",
    "34e09817-6cbe-4d01-b1a2-e0eac5743d41": "Kubernetes Cluster - Azure Arc Onboarding",
    "350f8d15-c687-4448-8ae1-157740a3936d": "Hierarchy Settings Administrator",
    "352470b3-6a9c-4686-b503-35deb827e500": "Security Detonation Chamber Publisher",
    "361898ef-9ed1-48c2-849c-a832951106bb": "Domain Services Reader",
    "36243c78-bf99-498c-9df9-86d9f8d28608": "Resource Policy Contributor",
    "36e80216-a7e8-4f42-a7e1-f12c98cbaf8a": "Impact Reporter",
    "3913510d-42f4-4e42-8a64-420c390055eb": "Monitoring Metrics Publisher",
    "392ae280-861d-42bd-9ea5-08ee6d83b80e": "Template Spec Reader",
    "39bc4728-0917-49c7-9d2c-d95423bc2eb4": "Security Reader",
    "3ae3fb29-0000-4ccd-bf80-542e7b26e081": "Load Test Reader",
    "3b20f47b-3825-43cb-8114-4bd2201156a8": "Cognitive Services Metrics Advisor User",
    "3d55a8f6-4133-418d-8051-facdb1735758": "Windows365SubscriptionReader",
    "3db33094-8700-4567-8da5-1501d4e7e843": "FHIR Data Exporter",
    "3df8b902-2a6f-47c7-8cc5-360e9b272a7e": "Remote Rendering Administrator",
    "3e150937-b8fe-4cfb-8069-0eaf05ecd056": "Microsoft Sentinel Responder",
    "3e5e47e6-65f7-47ef-90b5-e5dd4d455f24": "Disk Backup Reader",
    "3f2eb865-5811-4578-b90a-6fc6fa0df8e5": "Azure Front Door Secret Contributor",
    "3f88fce4-5892-4214-ae73-ba5294559913": "FHIR Data Writer",
    "40c5ff49-9181-41f8-ae61-143b0e78555e": "Desktop Virtualization Power On Off Contributor",
    "41077137-e803-4205-871c-5a86e6a753b4": "Blueprint Contributor",
    "420fcaa2-552c-430f-98ca-3264be4806c7": "SignalR App Server",
    "423170ca-a8f6-4b0f-8487-9e4eb8f49bfa": "Azure Maps Data Reader",
    "426e0c7f-0c7e-4658-b36f-ff54d6c29b45": "CDN Endpoint Contributor",
    "428e0ff0-5e57-4d9c-a221-2c70d0e0a443": "EventGrid EventSubscription Contributor",
    "4339b7cf-9826-4e41-b4ed-c7f4505dac08": "Code Signing Identity Verifier",
    "434105ed-43f6-45c7-a02f-909b2ba83430": "Cost Management Contributor",
    "434fb43a-c01c-447e-9f67-c3ad923cfaba": "Azure Kubernetes Fleet Manager RBAC Admin",
    "437d2ced-4a38-4302-8479-ed2bcb43d090": "Blueprint Operator",
    "43d0d8ad-25c7-4714-9337-8ba259a9fe05": "Monitoring Reader",
    "4447db05-44ed-4da3-ae60-6cbece780e32": "Chamber User",
    "4465e953-8ced-4406-a58e-0f6e3f3b530b": "FHIR Data Importer",
    "45bb0b16-2f0c-4e78-afaa-a07599b003f6": "Load Test Owner",
    "45d50f46-0b78-4001-a660-4198cbe8cd05": "DevCenter Dev Box User",
    "4633458b-17de-408a-b874-0445c86b69e6": "Key Vault Secrets User",
    "466ccd10-b268-4a11-b098-b4849f024126": "Cognitive Services QnA Maker Reader",
    "47b7735b-770e-4598-a7da-8b91488b4c88": "Data Lake Analytics Developer",
    "489581de-a3bd-480d-9518-53dea7416b33": "Desktop Virtualization Power On Contributor",
    "48b40c6e-82e0-4eb3-90d5-19e40f49b624": "Hybrid Server Resource Administrator",
    "4939a1f6-9ae0-4e48-a1e0-f2cbe897382d": "SQL Managed Instance Contributor",
    "494ae006-db33-4328-bf46-533a6560a3ca": "Site Recovery Operator",
    "494bdba2-168f-4f31-a0a1-191d2f7c028c": "IoT Hub Twin Contributor",
    "49632ef5-d9ac-41f4-b8e7-bbe587fa74a1": "Experimentation Reader",
    "49a72310-ab8d-41df-bbb0-79b649203868": "Desktop Virtualization Reader",
    "49e2f5d2-7741-4835-8efa-19e1fe35e47f": "Device Update Deployments Reader",
    "4a167cdf-cb95-4554-9203-2347fe489bd9": "Object Anchors Account Reader",
    "4a9ae827-6dc8-4573-8ac7-8239d42aa03f": "Tag Contributor",
    "4abbcc35-e782-43d8-92c5-2d3f1bd2253f": "Azure Kubernetes Service Cluster User Role",
    "4ba50f17-9666-485c-a643-ff00808643f0": "FHIR SMART User",
    "4c8d0bbc-75d3-4935-991f-5f3c56d81508": "FHIR Data Reader",
    "4d97b98b-1d4f-4787-a291-c67834d212e7": "Network Contributor",
    "4dd61c23-6743-42fe-a388-d8bdd41cb745": "Object Understanding Account Owner",
    "4e9b8407-af2e-495b-ae54-bb60a55b1b5a": "Chamber Admin",
    "4ea46cd5-c1b2-4a8e-910b-273211f9ce47": "IoT Hub Registry Contributor",
    "4f6d3b9b-027b-4f4c-9142-0e5a2a2247e0": "Azure Service Bus Data Receiver",
    "4f8fab4f-1852-4a58-a46a-8eaf358af14a": "Avere Contributor",
    "4fc6c259-987e-4a07-842e-c321cc9d413f": "IoT Hub Data Contributor",
    "4fe576fe-1146-4730-92eb-48519fa6bf9f": "Automation Job Operator",
    "515c2055-d9d4-4321-b1b9-bd0c9a0f79fe": "Logic App Operator",
    "516239f1-63e1-4d78-a4de-a74fb236a071": "App Configuration Data Reader",
    "51d6186e-6489-4900-b93f-92e23144cca5": "Microsoft Sentinel Playbook Operator",
    "532bc159-b25e-42c0-969e-a1d439f60d77": "Media Services Live Events Administrator",
    "5432c526-bc82-444a-b7ba-57c5b0b5b34f": "CosmosRestoreOperator",
    "5548b2cf-c94c-4228-90ba-30851930a12f": "Microsoft.Kubernetes connected cluster role",
    "56328988-075d-4c6a-8766-d93edd6725b6": "API Management Workspace API Developer",
    "58a3b984-7adf-4c20-983a-32417c86fbc8": "DICOM Data Owner",
    "5a1fc7df-4bf1-4951-a576-89034ee01acd": "FHIR Data Contributor",
    "5ae67dd6-50cb-40e7-96ff-dc2bfa4b606b": "App Configuration Data Owner",
    "5af6afb3-c06c-4fa4-8848-71a8aee05683": "Azure Kubernetes Fleet Manager RBAC Writer",
    "5b999177-9696-4545-85c7-50de3797e5a1": "Azure Arc Kubernetes Writer",
    "5bd9cd88-fe45-4216-938b-f97437e15450": "DocumentDB Account Contributor",
    "5c4089e1-6d96-4d2f-b296-c1bc7137275f": "Cognitive Services Custom Vision Deployment",
    "5d1e5ee4-7c68-4a71-ac8b-0739630a3dfb": "Hybrid Server Onboarding",
    "5d28c62d-5b37-4476-8438-e587778df237": "New Relic APM Account Contributor",
    "5d51204f-eb77-4b1c-b86a-2ec626c49413": "Spatial Anchors Account Reader",
    "5d58bcaf-24a5-4b20-bdb6-eed9f69fbe4c": "Management Group Contributor",
    "5daaa2af-1fe8-407c-9122-bba179798270": "Lab Contributor",
    "5dffeca3-4936-4216-b2bc-10343a5abb25": "Schema Registry Contributor (Preview)",
    "5e0bd9bd-7b93-4f28-af87-19fc36ad61bd": "Cognitive Services OpenAI User",
    "5e28a61e-8040-49db-b175-bb5b88af6239": "Community Owner Role",
    "5e3c6656-6cfa-4708-81fe-0de47ac73342": "BizTalk Contributor",
    "5e467623-bb1f-42f4-a55d-6e525e11384b": "Backup Contributor",
    "5fb5aef8-1081-4b8e-bb16-9d5d0385bab5": "Automation Runbook Operator",
    "602da2ba-a5c2-41da-b01d-5360126ab525": "Virtual Machine Local User Login",
    "60921a7e-fef1-4a43-9b16-a26c52ad4769": "Grafana Viewer",
    "60fc6e62-5479-42d4-8bf4-67625fcc2840": "Disk Pool Operator",
    "612c2aa1-cb24-443b-ac28-3ab7272de6f5": "Security Assessment Contributor",
    "6188b7c9-7d01-4f99-a59f-c88b630326c0": "Experimentation Metric Contributor",
    "61ed4efc-fab3-44fd-b111-e24485cc132a": "HDInsight Cluster Operator",
    "6322a993-d5c9-4bed-b113-e49bbea25b27": "Cognitive Services LUIS Writer",
    "635dd51f-9968-44d3-b7fb-6d9a6bd613ae": "AzureML Metrics Writer (preview)",
    "63bb64ad-9799-4770-b5c3-24ed299a07bf": "Azure Kubernetes Fleet Manager Contributor Role",
    "63f0a09d-1495-4db4-a681-037d84835eb4": "Azure Arc Kubernetes Viewer",
    "641177b8-a67a-45b9-a033-47bc880bb21e": "Managed Application Contributor Role",
    "6670b86e-a3f7-4917-ac9b-5d6ab1be4567": "Site Recovery Contributor",
    "66bb4e9e-b016-4a94-8249-4c0511c2be84": "Policy Insights Data Writer (Preview)",
    "673868aa-7521-48a0-acc6-0f60742d39f5": "Data Factory Contributor",
    "67d33e57-3129-45e6-bb0b-7cc522f762fa": "Azure Arc VMware Private Clouds Onboarding",
    "68ff5d27-c7f5-4fa9-a21c-785d0df7bd9e": "Impact Reader",
    "69566ab7-960f-475b-8e7c-b3118f30c6bd": "Storage File Data Privileged Contributor",
    "69a216fc-b8fb-44d8-bc22-1f3c2cd27a39": "Azure Service Bus Data Sender",
    "6aac74c4-6311-40d2-bbdd-7d01e7c6e3a9": "Azure Arc ScVmm Private Clouds Onboarding",
    "6ae96244-5829-4925-a7d3-5975537d91dd": "Azure VM Managed identities restore Contributor",
    "6b77f0a0-0d89-41cc-acd1-579c22c17a67": "AgFood Platform Sensor Partner Contributor",
    "6be48352-4f82-47c9-ad5e-0acacefdb005": "Azure Maps Search and Render Data Reader",
    "6cef56e8-d556-48e5-a04f-b8e64114680f": "AcrImageSigner",
    "6d8ee4ec-f05a-4a1d-8b00-a9b17e38b437": "SQL Server Contributor",
    "6d949e1d-41e2-46e3-8920-c6e4f31a8310": "Azure Center for SAP solutions Management role",
    "6d994134-994b-4a59-9974-f479f0b227fb": "Azure Sphere Publisher",
    "6f12a6df-dd06-4f3e-bcb1-ce8be600526a": "Azure Stack Registration Owner",
    "70bbe301-9835-447d-afdd-19eb3167307c": "Spatial Anchors Account Owner",
    "71522526-b88f-4d52-b57f-d31fc3546d0d": "API Management Service Reader Role",
    "72fafb9e-0641-4937-9268-a91bfd8191a3": "Cost Management Reader",
    "7392c568-9289-4bde-aaaa-b7131215889d": "Azure Extension for SQL Server Deployment",
    "73c2c328-d004-4c5e-938c-35c6f5679a1f": "API Management Workspace API Product Manager",
    "73c42c96-874c-492b-b04d-ab87d138a893": "Log Analytics Reader",
    "749a398d-560b-491b-bb21-08924219302e": "Load Test Contributor",
    "749f88d5-cbae-40b8-bcfc-e573ddc772fa": "Monitoring Contributor",
    "76199698-9eea-4c19-bc75-cec21354c6b6": "Storage Table Data Reader",
    "76283e04-6283-4c54-8f91-bcf1374a3c64": "DevTest Labs User",
    "7628b7b8-a8b2-4cdc-b46f-e9b35248918e": "Cognitive Services Language Reader",
    "76cc9ee4-d5d3-4a45-a930-26add3d73475": "Access Review Operator Service Role",
    "79b01272-bf9f-4f4c-9517-5506269cf524": "Cognitive Search Serverless Data Reader",
    "7a6f0e70-c033-4fb1-828c-08514e5f4102": "Collaborative Runtime Operator",
    "7ac06ca7-21ca-47e3-a67b-cbd6e6223baf": "Cognitive Search Serverless Data Contributor",
    "7b0c7e81-271f-4c71-90bf-e30bdfdbc2f7": "Azure Center for SAP solutions administrator",
    "7ca78c08-252a-4471-8644-bb5ff32d4ba0": "Search Service Contributor",
    "7e4f1700-ea5a-4f59-8f37-079cfe29dce3": "SignalR Service Owner",
    "7eabc9a4-85f7-4f71-b8ab-75daaccc1033": "Windows365NetworkUser",
    "7ec7ccdc-f61e-41fe-9aaf-980df0a44eba": "AgFood Platform Service Reader",
    "7efff54f-a5b4-42b5-a1c5-5411624893ce": "Disk Snapshot Contributor",
    "7f646f1b-fa08-80eb-a22b-edd6ce5c915c": "Experimentation Contributor",
    "7f646f1b-fa08-80eb-a33b-edd6ce5c915c": "Experimentation Administrator",
    "7f6c6a51-bcf8-42ba-9220-52d62157d7db": "Azure Kubernetes Service RBAC Reader",
    "7f951dda-4ed3-4680-a7ca-43fe172d538d": "AcrPull",
    "80558df3-64f9-4c0f-b32d-e5094b036b0b": "Azure Spring Apps Connect Role",
    "80dcbedb-47ef-405d-95bd-188a1b4ac406": "Elastic SAN Owner",
    "81a9662b-bebf-436f-a333-f67b29880f12": "Storage Account Key Operator Service Role",
    "82200a5b-e217-47a5-b665-6d8765ee745b": "Services Hub Operator",
    "8311e382-0749-4cb8-b61a-304f252e45ec": "AcrPush",
    "8393591c-06b9-48a2-a542-1bd6b377f6a2": "Azure Arc Kubernetes Cluster Admin",
    "8508508a-4469-4e45-963b-2518ee0bb728": "AgFood Platform Service Contributor",
    "85cb6faf-e071-4c9b-8136-154b5a04f717": "Kubernetes Extension Contributor",
    "86240b0e-9422-4c43-887b-b61143f32ba8": "Desktop Virtualization Application Group Contributor",
    "86e8f5dc-a6e9-4c67-9d15-de283e8eac25": "Classic Storage Account Contributor",
    "871e35f6-b5c1-49cc-a043-bde969a0f2cd": "CDN Endpoint Reader",
    "87a39d53-fc1b-424a-814c-f7e04687dc9e": "Logic App Contributor",
    "88424f51-ebe7-446f-bc41-7fa16989e96c": "Cognitive Services Custom Vision Labeler",
    "8a0f0c08-91a1-4084-bc3d-661d67233fed": "Storage Queue Data Message Processor",
    "8a3c2885-9b38-4fd2-9d99-91af537c1347": "Purview role 1 (Deprecated)",
    "8b9dfcab-4b77-4632-a6df-94bd07820648": "Azure Sphere Contributor",
    "8bbe83f1-e2a6-4df7-8cb4-4e04d4e5c827": "Spatial Anchors Account Contributor",
    "8cf5e20a-e4b2-4e9d-b3a1-5ceb692c2761": "SignalR/Web PubSub Contributor",
    "8d289c81-5878-46d4-8554-54e1e3d8b5cb": "Microsoft Sentinel Reader",
    "8d8d5a11-05d3-4bda-a417-a08778121c7c": "HDInsight Domain Services Contributor",
    "8e3af657-a8ff-443c-a75c-2fe8c4bcb635": "Owner",
    "8ebe5a00-799e-43f5-93ac-243d3dce84a7": "Search Index Data Contributor",
    "8f5e0ce6-4f7b-4dcf-bddf-e6f48634a204": "Azure Maps Data Contributor",
    "8f96442b-4075-438f-813d-ad51ab4019af": "CDN Profile Reader",
    "9106cda0-8a86-4e81-b686-29a22c54effe": "ClearDB MySQL DB Contributor",
    "91c1777a-f3dc-4fae-b103-61d183457e46": "Managed Services Registration assignment Delete Role",
    "92aaf0da-9dab-42b6-94a3-d43ce8d16293": "Log Analytics Contributor",
    "93586559-c37d-4a6b-ba08-b9f0940c2d73": "Cognitive Services Custom Vision Reader",
    "9565a273-41b9-4368-97d2-aeb0c976a9b3": "API Management Service Workspace API Developer",
    "959f8984-c045-4866-89c7-12bf9737be2e": "Data Operator for Managed Disks",
    "974c5e8b-45b9-4653-ba55-5f855dd0fb88": "Storage Queue Data Contributor",
    "985d6b00-f706-48f5-a6fe-d0ca12fb668d": "Classic Storage Account Key Operator Service Role",
    "9894cab4-e18a-44aa-828b-cb588cd6f2d7": "Cognitive Services Face Recognizer",
    "9980e02c-c2be-4d73-94e8-173b1dc7cf3c": "Virtual Machine Contributor",
    "99dba123-b5fe-44d5-874c-ced7199a5804": "Media Services Streaming Endpoints Administrator",
    "9b7fa17d-e63e-47b0-bb0a-15c516ac86ec": "SQL DB Contributor",
    "9ef4ef9c-a049-46b0-82ab-dd8ac094c889": "Project Babylon Data Curator",
    "a001fd3d-188f-4b5d-821b-7da978bf7442": "Cognitive Services OpenAI Contributor",
    "a06f5c24-21a7-4e1a-aa2b-f19eb6684f5b": "Azure Spring Cloud Config Server Contributor",
    "a1705bd2-3a8f-45a5-8683-466fcfd5cc24": "FHIR Data Converter",
    "a2138dac-4907-4679-a376-736901ed8ad8": "AnyBuild Builder",
    "a2c4a527-7dc0-4ee3-897b-403ade70fafb": "Video Indexer Restricted Viewer",
    "a36e6959-b6be-4b12-8e9f-ef4b474d304d": "Lab Operator",
    "a37b566d-3efa-4beb-a2f2-698963fa42ce": "Security Detonation Chamber Submission Manager",
    "a41e2c5b-bd99-4a07-88f4-9bf657a760b8": "Integration Service Environment Contributor",
    "a4417e6f-fecd-4de8-b567-7b0420556985": "Key Vault Certificates Officer",
    "a4b10055-b0c7-44c2-b00f-c7b5b3550cf7": "Traffic Manager Contributor",
    "a6333a3e-0164-44c3-b281-7a577aff287f": "Windows Admin Center Administrator Login",
    "a638d3c7-ab3a-418d-83e6-5f17a39d4fde": "Azure Event Hubs Data Receiver",
    "a7264617-510b-434b-a828-9731dc254ea7": "Storage File Data SMB Share Elevated Contributor",
    "a795c7a0-d4a2-40c1-ae25-d81f01202912": "Backup Reader",
    "a79a5197-3a5c-4973-a920-486035ffd60f": "Grafana Editor",
    "a7ffa36f-339b-4b5c-8bdf-e2c188b2c0eb": "Azure Kubernetes Service RBAC Writer",
    "a8281131-f312-4f34-8d98-ae12be9f0d23": "Elastic SAN Volume Group Owner",
    "a8835c7d-b5cb-47fa-b6f0-65ea10ce07a2": "LocalNGFirewallAdministrator role",
    "a92dfd61-77f9-4aec-a531-19858b406c87": "Azure Arc ScVmm Administrator role",
    "a959dbd1-f747-45e3-8ba6-dd80f235f97c": "Desktop Virtualization Virtual Machine Contributor",
    "a97b65f3-24c7-4388-baec-2e87135dc908": "Cognitive Services User",
    "a99b0159-1064-4c22-a57b-c9b3caa1c054": "Azure Spring Apps Remote Debugging Role",
    "a9a19cc5-31f4-447c-901f-56c0bb18fcaf": "PlayFab Reader",
    "a9b99099-ead7-47db-8fcf-072597a61dfa": "Bayer Ag Powered Services CWUM Solution User Role",
    "aabbc5dd-1af0-458b-a942-81af88f9c138": "Azure Center for SAP solutions service role",
    "ab8e14d6-4a74-4a29-9ba8-549422addade": "Microsoft Sentinel Contributor",
    "aba4ae5f-2193-4029-9191-0cb91df5e314": "Storage File Data SMB Share Reader",
    "ac63b705-f282-497d-ac71-919bf39d939d": "Management Group Reader",
    "acdd72a7-3385-48ef-bd42-f606fba81ae7": "Reader",
    "ad2dd5fb-cd4b-4fd4-a9b6-4fed3630980b": "ContainerApp Reader",
    "add466c9-e687-43fc-8d98-dfcf8d720be5": "Data Box Contributor",
    "ae349356-3a1b-4a5e-921d-050484c6347e": "Application Insights Component Contributor",
    "ae8036db-e102-405b-a1b9-bae082ea436d": "SqlVM Migration Role",
    "aebf23d0-b568-4e86-b8f9-fe83a2c6ab55": "Desktop Virtualization Application Group Reader",
    "af6a70f8-3c9f-4105-acf1-d719e9fca4ca": "Elastic SAN Reader",
    "b0d8363b-8ddd-447d-831f-62ca05bff136": "Monitoring Data Reader",
    "b12aa53e-6015-4669-85d0-8515ebb3ae7f": "Private DNS Zone Contributor",
    "b1ff04bb-8a4e-4dc4-8eb5-8693973ce19b": "Azure Kubernetes Service RBAC Cluster Admin",
    "b24988ac-6180-42a0-ab88-20f7382dd24c": "Contributor",
    "b279062a-9be3-42a0-92ae-8b3cf002ec4d": "Workbook Reader",
    "b2de6794-95db-4659-8781-7e080d3f2b9d": "Cognitive Services Immersive Reader User",
    "b34d265f-36f7-4a0d-a4d4-e158ca92e90f": "Classic Network Contributor",
    "b447c946-2db7-41ec-983d-d8bf3b1c77e3": "IoT Hub Data Reader",
    "b50d9833-a0cb-478e-945f-707fcc997c13": "Disk Restore Operator",
    "b5537268-8956-4941-a8f0-646150406f0c": "Azure Spring Cloud Data Reader",
    "b59867f0-fa02-499b-be73-45a86b5b3e1c": "Cognitive Services Data Reader (Preview)",
    "b60367af-1334-4454-b71e-769d9a4f83d9": "Graph Owner",
    "b64e21ea-ac4e-4cdf-9dc9-5b892992bee7": "Azure Connected Machine Onboarding",
    "b748a06d-6150-4f8a-aaa9-ce3940cd96cb": "Azure Arc VMware VM Contributor",
    "b7e6dc6d-f1e8-4753-8033-0f276bb0955b": "Storage Blob Data Owner",
    "b86a8fe4-44ce-4948-aee5-eccb2c155cd7": "Key Vault Secrets Officer",
    "b8b15564-4fa6-4a59-ab12-03e1d9594795": "Autonomous Development Platform Data Contributor (Preview)",
    "b8eda974-7b85-4f76-af95-65846b26df6d": "Storage File Data Privileged Reader",
    "b9331d33-8a36-4f8c-b097-4f54124fdb44": "Managed Applications Reader",
    "b97fb8bc-a8b2-4522-a38b-dd33c7e65ead": "Lab Creator",
    "ba79058c-0414-4a34-9e42-c3399d80cd5a": "Kubernetes Namespace User",
    "ba92f5b4-2d11-453d-a403-e96b0029c9fe": "Storage Blob Data Contributor",
    "bbf86eb8-f7b4-4cce-96e4-18cddf81d86e": "Attestation Contributor",
    "bcd981a7-7f74-457b-83e1-cceb9e632ffe": "Azure Digital Twins Data Owner",
    "bda0d508-adf1-4af0-9c28-88919fc3ae06": "Azure Stack HCI registration role",
    "befefa01-2a29-4197-83a8-272ff33ce314": "DNS Zone Contributor",
    "bfb1c7d2-fb1a-466b-b2ba-aee63b92deaf": "Web PubSub Service Reader",
    "bfc3b73d-c6ff-45eb-9a5f-40298295bf20": "LocalRulestacksAdministrator role",
    "c025889f-8102-4ebf-b32c-fc0c6f0c6bd9": "Avere Operator",
    "c031e6a8-4391-4de0-8d69-4706a7ed3729": "API Management Developer Portal Content Editor",
    "c0781e91-8102-4553-8951-97c6d4243cda": "Azure Arc ScVmm Private Cloud User",
    "c12c1c16-33a1-487b-954d-41c89c60f349": "Reader and Data Access",
    "c1ff6cc2-c111-46fe-8896-e0ef812ad9f3": "Cognitive Services Custom Vision Contributor",
    "c2f4ef07-c644-48eb-af81-4b1b4947fb11": "AcrDelete",
    "c4bba371-dacd-4a26-b320-7250bca963ae": "Media Services Policy Administrator",
    "c4bc862a-3b64-4a35-a021-a380c159b042": "Bayer Ag Powered Services GDU Solution",
    "c6a89b2d-59bc-44d0-9896-0f6e12d7b80a": "Storage Queue Data Message Sender",
    "c6decf44-fd0a-444c-a844-d653c394e7ab": "Data Labeling - Labeler",
    "c7393b34-138c-406f-901b-d8cf2b17e6ae": "Managed Application Operator Role",
    "c7aa55d3-1abb-444a-a5ca-5e51e485d6ec": "Integration Service Environment Developer",
    "c8ae6279-5a0b-4cb2-b3f0-d4d62845742c": "Azure Sphere Reader",
    "c8d4ff99-41c3-41a8-9f60-21dfdad59608": "AcrQuarantineWriter",
    "c8d896ba-346d-4f50-bc1d-7d1c84130446": "Project Babylon Data Reader",
    "ca0835dd-bacc-42dd-8ed2-ed5e7230d15b": "Object Anchors Account Owner",
    "ca6382a4-1721-4bcf-a114-ff0c70227b6b": "Application Group Contributor",
    "cb43c632-a144-4ec5-977c-e80c4affc34a": "Cognitive Services Metrics Advisor Administrator",
    "cd08ab90-6b14-449c-ad9a-8f8e549482c6": "Scheduled Patching Contributor",
    "cd570a14-e51a-42ad-bac8-bafd67325302": "Azure Connected Machine Resource Administrator",
    "cdda3590-29a3-44f6-95f2-9f980659eb04": "AcrQuarantineReader",
    "ce40b423-cede-4313-a93f-9b28290b72e1": "Lab Assistant",
    "ce551c02-7c42-47e0-9deb-e3b6fc3a9a83": "Azure Arc VMware Private Cloud User",
    "ceadfde2-b300-400a-ab7b-6143895aa822": "Desktop Virtualization Host Pool Reader",
    "cfd33db0-3dd1-45e3-aa9d-cdbdf3b6f24e": "Support Request Contributor",
    "cff1b556-2399-4e7e-856d-a8f754be7b65": "Azure Spring Cloud Service Registry Reader",
    "d04c6db6-4947-4782-9e91-30a88feb7be7": "Azure Spring Cloud Config Server Reader",
    "d17ce0a2-0697-43bc-aac5-9113337ab61c": "WorkloadBuilder Migration Agent Role",
    "d18777c0-1514-4662-8490-608db7d334b6": "Object Understanding Account Reader",
    "d18ad5f3-1baf-4119-b49b-d944edb1f9d0": "MySQL Backup And Export Operator",
    "d1ee9a80-8b14-47f0-bdc2-f4a351625a7b": "Device Update Content Reader",
    "d24ecba3-c1f4-40fa-a7bb-4588a071e8fd": "VM Scanner Operator",
    "d3881f73-407a-4167-8283-e981cbba0404": "Automation Operator",
    "d39065c4-c120-43c9-ab0a-63eed9795f0a": "Remote Rendering Client",
    "d57506d4-4c8d-48b1-8587-93c323f6a5a3": "Azure Digital Twins Data Reader",
    "d59a3e9c-6d52-4a5a-aeed-6bf3cf0e31da": "API Management Service Workspace API Product Manager",
    "d5a2ae44-610b-4500-93be-660a0c5f5ca6": "Kubernetes Agentless Operator",
    "d5a91429-5739-47e2-a06b-3470a27159e7": "EventGrid Data Sender",
    "d63b75f7-47ea-4f27-92ac-e0d173aaf093": "Autonomous Development Platform Data Reader (Preview)",
    "d6470a16-71bd-43ab-86b3-6f3a73f4e787": "Azure Maps Data Read and Batch Role",
    "d73bb868-a0df-4d4d-bd69-98a00b01fccb": "Classic Virtual Machine Contributor",
    "daa9e50b-21df-454c-94a6-a8050adab352": "Collaborative Data Contributor",
    "db58b8e5-c6ad-4a2a-8342-4190687cbf4a": "Storage Blob Delegator",
    "db7b14f2-5adf-42da-9f96-f2ee17bab5cb": "CosmosBackupOperator",
    "dba33070-676a-4fb0-87fa-064dc56ff7fb": "Azure Maps Contributor",
    "dbaa88c4-0c30-4179-9fb3-46319faa6149": "Site Recovery Reader",
    "ddc140ed-e463-4246-9145-7c664192013f": "Azure Arc VMware Administrator role ",
    "ddde6b66-c0df-4114-a159-3618637b3035": "SignalR REST API Reader",
    "de139f84-1756-47ae-9be6-808fbbe84772": "Website Contributor",
    "dfce44e4-17b7-4bd1-a6d1-04996ec95633": "Device Provisioning Service Data Contributor",
    "dffb1e0c-446f-4dde-a09f-99eb5cc68b96": "Azure Arc Kubernetes Admin",
    "e022efe7-f5ba-4159-bbe4-b44f577e9b61": "API Management Service Operator Role",
    "e0f68234-74aa-48ed-b826-c38b57376e17": "Redis Cache Contributor",
    "e147488a-f6f5-4113-8e2d-b22465e65bf6": "Key Vault Crypto Service Encryption User",
    "e307426c-f9b6-4e81-87de-d99efb3c32bc": "Desktop Virtualization Host Pool Contributor",
    "e3d13bf0-dd5a-482e-ba6b-9b8433878d10": "Security Manager (Legacy)",
    "e40ec5ca-96e0-45a2-b4ff-59039f2c2b59": "Managed Identity Contributor",
    "e4237640-0e3d-4a46-8fda-70bc94856432": "Device Update Deployments Administrator",
    "e4395492-1534-4db2-bedf-88c14621589c": "Media Services Media Operator",
    "e503ece1-11d0-4e8e-8e2c-7a6c3bf38815": "AzureML Compute Operator",
    "e582369a-e17b-42a5-b10c-874c387c530b": "Azure Arc ScVmm VM Contributor",
    "e5e2a7ff-d759-4cd2-bb51-3152d37e2eb1": "Storage Account Backup Contributor",
    "e8113dce-c529-4d33-91fa-e9b972617508": "Azure Connected SQL Server Onboarding",
    "e89c7a3c-2f64-4fa1-a847-3e4c9ba4283a": "DICOM Data Reader",
    "e8ddcd69-c73f-4f9f-9844-4100522f16ad": "Workbook Contributor",
    "e9b8712a-cbcf-4ea7-b0f7-e71b803401e6": "SaaS Hub Contributor",
    "e9dba6fb-3d52-4cf0-bce3-f06ce71b9e0f": "Device Update Reader",
    "ea4bfff8-7fb4-485a-aadd-d4129a0ffaa6": "Desktop Virtualization User Session Operator",
    "ec156ff8-a8d1-4d15-830c-5b80698ca432": "CDN Profile Contributor",
    "ed7f3fbd-7b88-4dd4-9017-9adb7ce333f8": "Azure Kubernetes Service Contributor Role",
    "ee361c5d-f7b5-4119-b4b6-892157c8f64c": "Knowledge Consumer",
    "eeaeda52-9324-47f6-8069-5d5bade478b2": "Domain Services Contributor",
    "ef1c2c96-4a77-49e8-b9a4-6179fe1d2fd2": "API Management Workspace Reader",
    "ef29765d-0d37-4119-a4f8-f9f9902c9588": "Bayer Ag Powered Services Imagery Solution",
    "f0310ce6-e953-4cf8-b892-fb1c87eaf7f6": "Azure Usage Billing Data Sender",
    "f07febfe-79bc-46b1-8b37-790e26e6e498": "Cognitive Services Language Owner",
    "f1a07417-d97a-45cb-824c-7a7467783830": "Managed Identity Operator",
    "f2310ca1-dc64-4889-bb49-c8e0fa3d47a8": "Cognitive Services Language Writer",
    "f25e0fa2-a7c8-4377-a976-54943a77a395": "Key Vault Contributor",
    "f2dc8367-1007-4938-bd23-fe263f013447": "Cognitive Services Speech User",
    "f353d9bd-d4a6-484e-a77a-8050b599b867": "Automation Contributor",
    "f4c81013-99ee-4d62-a7ee-b3f1f648599a": "Microsoft Sentinel Automation Contributor",
    "f4cc2bf9-21be-47a1-bdf1-5c5804381025": "Cognitive Services QnA Maker Editor",
    "f526a384-b230-433a-b45c-95f59c4a2dec": "Azure Event Hubs Data Owner",
    "f5819b54-e033-4d82-ac66-4fec3cbf3f4c": "Azure Connected Machine Resource Manager",
    "f58310d9-a9f6-439a-9e8d-f62e7b41a168": "Role Based Access Control Administrator (Preview)",
    "f5880b48-c26d-48be-b172-7927bfa1c8f1": "Azure Spring Cloud Service Registry Contributor",
    "f69b8690-cc87-41d6-b77a-a4bc3c0a966f": "Lab Services Contributor",
    "f6c7c914-8db3-469d-8ca1-694a8f32e121": "AzureML Data Scientist",
    "f72c8140-2111-481c-87ff-72b910f6e3f8": "Cognitive Services LUIS Owner",
    "f7b75c60-3036-4b75-91c3-6b41c27c1689": "Reservation Purchaser",
    "f8da80de-1ff9-4747-ad80-a19b7f6079e3": "AgFood Platform Service Admin",
    "fa23ad8b-c56e-40d8-ac0c-ce449e1d2c64": "Billing Reader",
    "fb1c8493-542b-48eb-b624-b4c8fea62acd": "Security Admin",
    "fb879df8-f326-4884-b1cf-06f3ad86be52": "Virtual Machine User Login",
    "fbc52c3f-28ad-4303-a892-8a056630b8f1": "Azure Traffic Controller Configuration Manager",
    "fbdf93bf-df7d-467e-a4d2-9458aa1360c8": "Cosmos DB Account Reader Role",
    "fd1bd22b-8476-40bc-a0bc-69b95687b9f3": "Attestation Reader",
    "fd53cd77-2268-407a-8f46-7e7863d0f521": "SignalR REST API Owner",
    "ff100721-1b9d-43d8-af52-42b69c1272db": "Purview role 3 (Deprecated)",
    "ffc6bbe0-e443-4c3b-bf54-26581bb2f78e": "App Compliance Automation Reader"
}
